# Copyright 2018-2020 Nick Anthony, Backman Biophotonics Lab, Northwestern University
#
# This file is part of PWSpy.
#
# PWSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PWSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PWSpy.  If not, see <https://www.gnu.org/licenses/>.

from __future__ import annotations

import sys

from PyQt5 import QtCore
from PyQt5.QtWidgets import QMessageBox, QScrollArea, QWidget, QVBoxLayout, QLabel, QPushButton, QApplication
import typing


class ScrollableMessageBox(QMessageBox):
    def __init__(self, icon: QMessageBox.Icon, title: str, text: str, buttons: QMessageBox.StandardButtons = QMessageBox.NoButton, parent: QWidget = None,
                f: QtCore.Qt.WindowFlags = QtCore.Qt.Dialog | QtCore.Qt.MSWindowsFixedSizeDialogHint):
        QMessageBox.__init__(self, icon, title, "", buttons, parent, f)
        scroll = QScrollArea(self)
        scroll.setWidgetResizable(True)
        self.content = QWidget()
        lay = QVBoxLayout()
        self.content.setLayout(lay)
        scroll.setWidget(self.content)
        scroll.setStyleSheet("QScrollArea{min-width:300 px; min-height: 200px}")
        label = QLabel(text, self)
        label.setWordWrap(True)
        lay.addWidget(label)
        lay.setAlignment(QtCore.Qt.AlignTop)
        self.layout().replaceWidget(self.layout().itemAt(2).widget(), scroll)

    @staticmethod
    def question(parent: QWidget, title: str, text: str, buttons: typing.Union[QMessageBox.StandardButtons, QMessageBox.StandardButton] = QMessageBox.Yes|QMessageBox.No,
                 defaultButton: QMessageBox.StandardButton = QMessageBox.Yes) -> QMessageBox.StandardButton:
        scr = ScrollableMessageBox(QMessageBox.Question, title, text, buttons, parent)
        scr.setModal(True)
        scr.exec()
        return scr.result()

    @staticmethod
    def information(parent: QWidget, title: str, text: str, buttons: typing.Union[QMessageBox.StandardButtons, QMessageBox.StandardButton] = QMessageBox.Ok) -> QMessageBox.StandardButton:
        scr = ScrollableMessageBox(QMessageBox.Information, title, text, buttons, parent)
        scr.setModal(True)
        scr.exec()
        return scr.result()

if __name__ == "__main__":
    string = ""
    for i in range(10):
        string += str(i) + " "
    app = QApplication(sys.argv)
    button = QPushButton("GO")
    scr = ScrollableMessageBox.question(button, "Test Title", string)
    print(scr)
    sys.exit(app.exec())


from __future__ import annotations
import typing as t_
import warnings
import logging
from PyQt5 import QtCore
from PyQt5.QtWidgets import QWidget, QApplication, QInputDialog, QMessageBox
from pwspy_gui.PWSAnalysisApp.App import PWSApp

from pwspy_gui.PWSAnalysisApp.pluginInterfaces import CellSelectorPlugin
import os

from pwspy_gui.PWSAnalysisApp.plugins.acquisitionAwareROIDrawer._control import SequenceController

from ._ui import SeqRoiDrawer
from pwspy.utility.acquisition.sequencerCoordinate import SequencerCoordinateRange, SequenceAcquisition
from pwspy.utility.acquisition.steps import SequencerStep
from pwspy.utility.acquisition import RuntimeSequenceSettings
import pwspy.dataTypes as pwsdt
if t_.TYPE_CHECKING:
    from pwspy_gui.PWSAnalysisApp.componentInterfaces import CellSelector


class AcquisitionAwareRoiDrawerPlugin(CellSelectorPlugin):
    def __init__(self):
        self._selector: CellSelector = None
        self._parentWidget: QWidget = None
        self._ui: SeqRoiDrawer = None

    def setContext(self, selector: CellSelector, parent: QWidget):
        """set the CellSelector that this plugin is associated to."""
        self._selector = selector
        self._parentWidget = parent

    def onCellsSelected(self, cells: t_.List[pwsdt.Acquisition]):
        """This method will be called when the CellSelector indicates that it has had new cells selected."""
        pass

    def onReferenceSelected(self, cell: pwsdt.Acquisition):
        """This method will be called when the CellSelector indicates that it has had a new reference selected."""
        pass

    def onNewCellsLoaded(self, cells: t_.List[pwsdt.Acquisition]):
        """This method will be called when the CellSelector indicates that new cells have been loaded to the selector."""
        pass

    def getName(self) -> str:
        """The name to refer to this plugin by."""
        return "Sequence ROI Drawing"

    def onPluginSelected(self):
        """This method will be called when the plugin is activated."""
        cells = self._selector.getAllCellMetas()
        if len(cells) == 0:
            QMessageBox.information(self._parentWidget, "Hmm", "No images are loaded.")
            return
        sequence, acqs = self._loadFromAcqs(cells)

        anName, ok = QInputDialog.getText(self._parentWidget, 'Analysis Name', 'Please input the analysis name', text='p0')
        if not ok:  # Must have been cancelled.
            return
        controller = SequenceController(sequence, acqs)
        mds = []
        for sacq in acqs:
            acq = sacq.acquisition
            pwsAn = None
            dynAn = None
            if acq.pws is not None:
                if anName in acq.pws.getAnalyses():
                    pwsAn = acq.pws.loadAnalysis(anName)
            if acq.dynamics is not None:
                if anName in acq.dynamics.getAnalyses():
                    dynAn = acq.dynamics.loadAnalysis(anName)
            mds.append((sacq, (pwsAn, dynAn)))
        self._ui = SeqRoiDrawer(controller, mds, roiManager=PWSApp.instance().roiManager, parent=self._parentWidget)
        self._ui.show()

    def additionalColumnNames(self) -> t_.Sequence[str]:
        """The header names for each column."""
        return tuple()

    def getTableWidgets(self, acq: pwsdt.Acquisition) -> t_.Sequence[QWidget]:
        """provide a widget for each additional column to represent `acq`"""
        return tuple()


    def _loadFromAcqs(self, acqs: t_.Sequence[pwsdt.Acquisition]) -> t_.Tuple[SequencerStep, t_.Tuple[SequenceAcquisition, ...]]:
        commonPath = os.path.commonpath([acq.filePath for acq in acqs])
        logger = logging.getLogger(__name__)
        logger.debug(f"New cells loaded at common path: {commonPath}")
        for i in range(3):  # Look up to 3 parent directories for a valid sequence file.
            try:
                sequenceRoot = RuntimeSequenceSettings.fromJsonFile(commonPath)
                logger.debug(f"Loaded sequence file at {commonPath}")
                if sequenceRoot.uuid is None:
                    warnings.warn(
                        "Old acquisition sequence file must have been loaded. No UUID found. Acquisitions returned by this function may not actually belong to this sequence.")
            except FileNotFoundError:
                commonPath = os.path.split(commonPath)[0]  # Go up one directory
                logger.debug(f"No sequence file found. Trying again at: {commonPath}")
                continue

            foundAcqs = []  # We only get here if a sequence file was found.
            for f in acqs:
                try:
                    foundAcqs.append(SequenceAcquisition(f))
                except FileNotFoundError:
                    pass  # There may be "Cell" folders that don't contain a sequencer coordinate.
            foundAcqs = [acq for acq in foundAcqs if acq.sequencerCoordinate.uuid == sequenceRoot.uuid]  # Filter out acquisitions that don't have a matching UUID to the sequence file.
            logger.debug(f"Successfully loaded {len(foundAcqs)} sequenced acquisitions.")
            return sequenceRoot.rootStep, tuple(foundAcqs)



# Copyright 2018-2020 Nick Anthony, Backman Biophotonics Lab, Northwestern University
#
# This file is part of PWSpy.
#
# PWSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PWSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PWSpy.  If not, see <https://www.gnu.org/licenses/>.

import os

from PyQt5 import QtCore, QtGui
from PyQt5.QtWidgets import QMainWindow, QApplication, QToolBar, QMessageBox
from pwspy_gui.sharedWidgets.extraReflectionManager import ERManager
from . import resources
from pwspy_gui import resources as sharedresources
import pwspy
import pwspy_gui
from .componentInterfaces import CellSelector, AnalysisSettingsCreator, ResultsTableController, ROIManager
from .dialogs import WorkingDirDialog
from ._dockWidgets import CellSelectorDock, AnalysisSettingsDock, ResultsTableControllerDock, PlottingDock


class PWSWindow(QMainWindow):
    def __init__(self, erManager: ERManager, roiManager: ROIManager):
        super().__init__()
        self.setWindowTitle(QApplication.instance().applicationName())
        self.setWindowIcon(QtGui.QIcon(os.path.join(resources, 'cellLogo.png')))
        self.cellSelector: CellSelector = CellSelectorDock(parent=self, roiManager=roiManager)
        self.analysisSettings: AnalysisSettingsCreator = AnalysisSettingsDock(self, self.cellSelector, erManager)
        self.resultsTable: ResultsTableController = ResultsTableControllerDock(self)
        self.plots = PlottingDock(self.cellSelector)
        self.addDockWidget(QtCore.Qt.RightDockWidgetArea, self.cellSelector)
        self.addDockWidget(QtCore.Qt.LeftDockWidgetArea, self.plots)
        self.addDockWidget(QtCore.Qt.RightDockWidgetArea, self.analysisSettings)
        self.addDockWidget(QtCore.Qt.BottomDockWidgetArea, self.resultsTable)
        self.setDockOptions(QMainWindow.AnimatedDocks | QMainWindow.AllowNestedDocks | QMainWindow.AllowTabbedDocks)

        self.fileDialog = WorkingDirDialog(self)

        menuBar = self.menuBar()
        menu = menuBar.addMenu("View")
        defaultLayoutAction = menu.addAction("Set Default Layout")
        defaultLayoutAction.triggered.connect(self._setDefaultLayout)
        infoAction = menu.addAction("Info")
        infoAction.triggered.connect(self.openInfoPane)
        menu = menuBar.addMenu("Config")
        self.parallelAction = menu.addAction("Multi-Core Analysis (faster, needs more RAM)")
        self.parallelAction.setCheckable(True)
        menu = menuBar.addMenu("Actions")
        menu.setToolTipsVisible(True)
        self.blindAction = menu.addAction("Create blinded directory")
        self.blindAction.setToolTip("Creates a folder of symlinks to the selected data that is randomly numbered. "
                                    "This allows you to work on data anonymously without bias. You may need to run this "
                                    "software as `Admin` for this to work on Windows.")
        self.roiConvertAction = menu.addAction("Update ROI file formats")
        self.roiConvertAction.setToolTip("Updates old .MAT roiFile files to a newer .H5 format that will run more efficiently."
                                         " Warning: The old files will be deleted.")

        toolBar = QToolBar("Tool Bar", self)
        toolBar.setFloatable(False)
        toolBar.setObjectName('mainToolBar()')
        self.addToolBar(QtCore.Qt.LeftToolBarArea, toolBar)
        browseAction = toolBar.addAction(QtGui.QIcon(os.path.join(sharedresources, 'folder.svg')), "Set Path")
        browseAction.triggered.connect(self.fileDialog.show)
        self.runAction = toolBar.addAction(QtGui.QIcon(os.path.join(resources, 'playicon.svg')), 'Run').triggered
        settings = QtCore.QSettings("BackmanLab", "PWSAnalysis2")
        try:
            self.restoreGeometry(settings.value("geometry"))
            self.restoreState(settings.value("windowState"))
        except TypeError as e:  # Setting must not exist
            self.resize(1024, 768)
            self._setDefaultLayout()

    def closeEvent(self, event):
        settings = QtCore.QSettings("BackmanLab", "PWSAnalysis2")
        settings.setValue("geometry", self.saveGeometry())
        settings.setValue("windowState", self.saveState())
        self.cellSelector.close() #This causes the current cell items to save their metadata.
        super().closeEvent(event)

    def openInfoPane(self):
        msgBox = QMessageBox.information(self, "About PWS Analysis", f"This software is intended for the analysis of Partial Wave Spectroscopic microscopy data.\npwspy_gui version: {pwspy_gui.__version__}\npwspy version: {pwspy.__version__}")

    def _setDefaultLayout(self):
        #remove all docks then re add them
        docks = [self.cellSelector, self.plots, self.analysisSettings, self.resultsTable]
        for dock in docks:
            self.removeDockWidget(dock)
        self.addDockWidget(QtCore.Qt.RightDockWidgetArea, self.cellSelector)
        self.addDockWidget(QtCore.Qt.LeftDockWidgetArea, self.plots)
        #self.addDockWidget(QtCore.Qt.BottomDockWidgetArea, self.resultsTable)
        self.tabifyDockWidget(self.cellSelector, self.resultsTable)
        self.tabifyDockWidget(self.plots, self.analysisSettings)
        for dock in docks:
            dock.setFloating(False)
            dock.setVisible(True)



from __future__ import annotations
import typing
import warnings
import logging
from PyQt5 import QtCore
from PyQt5.QtWidgets import QWidget, QApplication

from pwspy_gui.PWSAnalysisApp.pluginInterfaces import CellSelectorPlugin
import os

from ._ui.widget import SequenceViewer
from pwspy.utility.acquisition.sequencerCoordinate import SequencerCoordinateRange, SequenceAcquisition
from pwspy.utility.acquisition.steps import SequencerStep
from pwspy.utility.acquisition import RuntimeSequenceSettings
from pwspy.dataTypes import Acquisition
if typing.TYPE_CHECKING:
    from pwspy_gui.PWSAnalysisApp.componentInterfaces import CellSelector


def requirePluginActive(method):
    """A decorator that only runs the decorated function if the plugin UI is open"""
    def newMethod(self, *args, **kwargs):
        if self._ui.isVisible():  # If the ui isn't visible then we consider the plugin to be off.
            method(self, *args, **kwargs)
    return newMethod


class AcquisitionSequencerPlugin(CellSelectorPlugin):
    def __init__(self):
        self._selector: CellSelector = None
        self._sequence: SequencerStep = None
        self._cells: typing.List[SequenceAcquisition] = None
        self._ui = SequenceViewer()
        self._ui.newCoordSelected.connect(self._updateSelectorSelection)

    def setContext(self, selector: CellSelector, parent: QWidget):
        """set the CellSelector that this plugin is associated to."""
        self._selector = selector
        self._ui.setParent(parent)
        self._ui.setWindowFlags(QtCore.Qt.Window)  # Without this is just gets added to the main window in a weird way.

    @requirePluginActive
    def onCellsSelected(self, cells: typing.List[pwsdt.Acquisition]):
        """This method will be called when the CellSelector indicates that it has had new cells selected."""
        pass

    @requirePluginActive
    def onReferenceSelected(self, cell: pwsdt.Acquisition):
        """This method will be called when the CellSelector indicates that it has had a new reference selected."""
        pass

    @requirePluginActive
    def onNewCellsLoaded(self, cells: typing.List[pwsdt.Acquisition]):
        """This method will be called when the CellSelector indicates that new cells have been loaded to the selector."""
        sequence, acqs = self._loadNewSequence(cells)
        self._sequence = sequence
        self._cells = acqs
        if sequence is not None:
            self._ui.setSequenceStepRoot(self._sequence)


    def getName(self) -> str:
        """The name to refer to this plugin by."""
        return "Acquisition Sequence Selector"

    def onPluginSelected(self):
        """This method will be called when the plugin is activated."""
        self._ui.show()  # We use ui visibility to determine if the plugin is active or not.
        self.onNewCellsLoaded(self._selector.getAllCellMetas())  # Make sure we're all up to date

    def additionalColumnNames(self) -> typing.Sequence[str]:
        """The header names for each column."""
        return tuple() #return "Coord. Type", "Coord. Value" # We used to add new columns, but it was confusing, better not to.

    def getTableWidgets(self, acq: pwsdt.Acquisition) -> typing.Sequence[QWidget]:  #TODO this gets called before the sequence has been loaded. Make it so this isn't required for constructor of cell table widgets.
        """provide a widget for each additional column to represent `acq`"""
        return tuple()
        # typeNames = {SequencerStepTypes.POS.name: "Position", SequencerStepTypes.TIME.name: "Time", SequencerStepTypes.ZSTACK.name: "Z Stack"}
        # try:
        #     acq = SequenceAcquisition(acq)
        # except:
        #     return tuple((QTableWidgetItem(), QTableWidgetItem()))
        # coord = acq.sequencerCoordinate
        # idx, iteration = [(i, iteration) for i, iteration in enumerate(coord.iterations) if iteration is not None][-1]
        # for step in self._sequence.iterateChildren():
        #     if step.id == coord.ids[idx]:
        #         step: CoordSequencerStep
        #         val = QTableWidgetItem(step.getIterationName(iteration))
        #         t = QTableWidgetItem(typeNames[step.stepType])
        #         return tuple((t, val))
        # return tuple((QTableWidgetItem(), QTableWidgetItem()))  # This will happen if the acquisition has a coords file but the coord isn't actually found in the sequence file.
        #

    def _updateSelectorSelection(self, coordRange: SequencerCoordinateRange):
        select: typing.List[Acquisition] = []
        for cell in self._cells:
            if cell.sequencerCoordinate in coordRange:
                select.append(cell.acquisition)
        self._selector.setSelectedCells(select)

    def _loadNewSequence(self, cells: typing.Sequence[pwsdt.Acquisition]) -> typing.Tuple[SequencerStep, typing.Sequence[SequenceAcquisition]]:
        if len(cells) == 0:  # This causes a crash
            return None, None
        cellFilePaths = [acq.filePath for acq in cells]

        #Search the parent directory for a `sequence.pwsseq` file containing the sequence information.
        paths = [i.filePath for i in cells]
        commonPath = os.path.commonpath(paths)
        logger = logging.getLogger(__name__)
        logger.debug(f"New cells loaded at common path: {commonPath}")
        # We will search up to 3 parent directories for a sequence file
        for i in range(3):
            try:
                sequenceRoot = RuntimeSequenceSettings.fromJsonFile(commonPath)
                if sequenceRoot.uuid is None:
                    warnings.warn("Old acquisition sequence file must have been loaded. No UUID found. Acquisitions returned by this function may not actually belong to this sequence.")
            except FileNotFoundError:
                commonPath = os.path.split(commonPath)[0]  # Go up one directory
                logger.debug(f"No sequence file found. Trying again at: {commonPath}")
                continue

            foundAcqs = []
            for f in cells:
                try:
                    foundAcqs.append(SequenceAcquisition(f))
                except FileNotFoundError:
                    pass  # There may be "Cell" folders that don't contain a sequencer coordinate.
            foundAcqs = [acq for acq in foundAcqs if acq.sequencerCoordinate.uuid == sequenceRoot.uuid]  # Filter out acquisitions that don't have a matching UUID to the sequence file.
            logger.debug(f"Successfully loaded {len(foundAcqs)} sequenced acquisitions.")
            foundAcqs = [acq for acq in foundAcqs if acq.acquisition.filePath in cellFilePaths]  # TODo what is the purpose of this
            sequence = sequenceRoot.rootStep
            return sequence, foundAcqs
        return None, None  # Nothing was found

if __name__ == '__main__':
    from pwspy_gui.PWSAnalysisApp.plugins.acquisitionSequencer._ui.TreeView import MyTreeView

    with open(r'C:\Users\nicke\Desktop\data\toast2\sequence.pwsseq') as f:
        s = SequencerStep.fromJson(f.read())
    import sys
    from pwspy import dataTypes as pwsdt
    from glob import glob

    acqs = [pwsdt.Acquisition(i) for i in glob(r"C:\Users\nicke\Desktop\data\toast2\Cell*")]
    sacqs = [SequenceAcquisition(acq) for acq in acqs]

    import sys

    app = QApplication(sys.argv)


    view = MyTreeView()
    view.setRoot(s)

    view.setWindowTitle("Simple Tree Model")
    view.show()
    sys.exit(app.exec_())


    app = QApplication(sys.argv)

    W = QWidget()
    W.setLayout(QHBoxLayout())

    w = QTreeWidget()
    w.setColumnCount(2)
    w.addTopLevelItem(s)
    w.setIndentation(10)

    w2 = DictTreeView()
    w2.setColumnCount(2)
    w2.setIndentation(10)


    w.itemClicked.connect(lambda item, column: w2.setDict(item.settings))

    W.layout().addWidget(w)
    W.layout().addWidget(w2)


    W.show()
    app.exec()
    a = 1



from __future__ import annotations
import typing as t_
from PyQt5 import QtCore
from PyQt5.QtWidgets import QWidget, QVBoxLayout, QHBoxLayout
from pwspy import dataTypes as pwsdt
from pwspy.utility.acquisition import SequenceAcquisition

from pwspy_gui.PWSAnalysisApp._roiManager import _DefaultROIManager
from pwspy_gui.PWSAnalysisApp.componentInterfaces import ROIManager
from pwspy_gui.PWSAnalysisApp.plugins.acquisitionAwareROIDrawer._control import SequenceController, RoiController, Options
from pwspy_gui.PWSAnalysisApp.plugins.acquisitionAwareROIDrawer._ui._buttonBar import ButtonBar
from pwspy_gui.PWSAnalysisApp.plugins.acquisitionAwareROIDrawer._ui._optionsPanel import OptionsPanel
from pwspy_gui.PWSAnalysisApp.sharedWidgets.plotting import RoiDrawer

if t_.TYPE_CHECKING:
    from pwspy.analysis.dynamics import DynamicsAnalysisResults
    from pwspy.analysis.pws import PWSAnalysisResults
    AnalysisResultsComboType = t_.Tuple[
        t_.Optional[PWSAnalysisResults],
        t_.Optional[DynamicsAnalysisResults]]


class SeqRoiDrawer(QWidget):
    def __init__(self, controller: SequenceController, metadatas: t_.List[t_.Tuple[SequenceAcquisition, t_.Optional[AnalysisResultsComboType]]], roiManager: ROIManager, parent: QWidget = None, flags=QtCore.Qt.Window):
        super().__init__(parent=parent, flags=flags)
        self.setWindowTitle("Sequence-Aware ROI Drawer")
        self._seqController = controller

        self._drawer = RoiDrawer(metadatas=[(seqAcq.acquisition, anResults) for seqAcq, anResults in metadatas], roiManager=roiManager, parent=self, flags=QtCore.Qt.Widget)  # Override the default behavior of showing as it's own window.
        self._roiController = RoiController(self._seqController, initialOptions=Options(False, False), roiManager=roiManager, parent=self)

        self._drawer.metadataChanged.connect(self._drawMetaDataChangeUnprompted)
        self._drawer.roiCreated.connect(lambda acq, roi, overwrite: self._roiController.setRoiChanged(acq, roi, overwrite))
        self._drawer.roiDeleted.connect(self._roiController.deleteRoi)
        self._drawer.roiModified.connect(lambda acq, roi: self._roiController.setRoiChanged(acq, roi, True))
        self._ignoreDrawerSignals = False

        self._optionsPanel = OptionsPanel(parent=self, initialOptions=[self._roiController.getOptions(i) for i in range(len(controller.iterSteps))])
        self._optionsPanel.animateTimerFired.connect(self._handleAnimationEvent)
        self._optionsPanel.optionsChanged.connect(self._roiController.setOptions)

        l = QVBoxLayout()

        self._buttonBars = []
        for step in controller.iterSteps:
            bBar = ButtonBar([step.getIterationName(i) for i in range(step.stepIterations())], self)
            bBar.buttonClicked.connect(self.setAcquisition)
            self._buttonBars.append(bBar)
            l.addWidget(bBar)

        ll = QHBoxLayout()
        ll.addWidget(self._drawer, stretch=1)  # Hog as much space as possible.
        ll.addWidget(self._optionsPanel)

        lll = QVBoxLayout(self)
        lll.addLayout(ll, stretch=1)
        lll.addLayout(l)
        self.setLayout(lll)

    def setAcquisition(self):
        idxs = tuple(bBar.getSelectedButtonId() for bBar in self._buttonBars)
        acq = self._seqController.setCoordinates(*idxs)
        self._ignoreDrawerSignals = True
        self._drawer.setDisplayedAcquisition(acq)
        self._ignoreDrawerSignals = False

    def _handleAnimationEvent(self, axis: int):
        self._buttonBars[axis].selectNextButton()

    def _drawMetaDataChangeUnprompted(self, acq: pwsdt.Acquisition):
        if not self._ignoreDrawerSignals:
            idxs = self._seqController.getIndicesForAcquisition(acq)
            for idx, bBar in zip(idxs, self._buttonBars):
                bBar.setButtonSelected(idx)


if __name__ == '__main__':
    from pwspy.utility.acquisition import loadDirectory
    from PyQt5.QtWidgets import QApplication
    import sys
    # d = r'\\BackmanLabNAS\home\Year3\KuriosBandwidth\data' # Positions only experiment
    d = r'\\BackmanLabNAS\Public\surbhi_nick_share\MCF10A D-Ala media auto'  # Time series and positions
    sequence, acqs = loadDirectory(d)
    cont = SequenceController(sequence, acqs)

    app = QApplication(sys.argv)
    roiManager = _DefaultROIManager()
    drawer = SeqRoiDrawer(cont, [(acq, (None, None)) for acq in acqs], roiManager=roiManager)
    drawer.show()
    app.exec()
    a = 1



# Copyright 2018-2020 Nick Anthony, Backman Biophotonics Lab, Northwestern University
#
# This file is part of PWSpy.
#
# PWSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PWSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PWSpy.  If not, see <https://www.gnu.org/licenses/>.

import copy
import logging
import os
from typing import List

from pwspy.dataTypes import Acquisition, Roi
from pwspy.dataTypes._other import RoiFile


class RoiConverter:
    """A class that converts old-style .mat roiFile files to the newer .h5 files.
    The key difference here is that the new files contain an array of vertices
    that specify the outline of the roiFile. Without these vertices they must be
    calculated using the concave hull method which is slow."""
    def __init__(self, cells: List[Acquisition]):
        for cell in cells:
            logger = logging.getLogger(__name__)
            logger.info(cell.filePath)
            rois = cell.getRois()
            for name, num, fformat in rois:
                if fformat == RoiFile.FileFormats.MAT:
                    logger.info('\t', name, num, "MAT")
                    roiFile = RoiFile.fromMat(cell.filePath, name, num)
                elif fformat == RoiFile.FileFormats.HDF:
                    logger.info('\t', name, num, "LegacyHDF")
                    roiFile = RoiFile.fromHDF_legacy(cell.filePath, name, num)
                else:
                    logger.info('\t', "Skipping", name, num, fformat.name)
                    continue #Conversion of other formats is not supported
                newFile = RoiFile.toHDF(roiFile.getRoi(), roiFile.name, roiFile.number, cell.filePath)  # save to Roi.FileFormat.HDF2 format. At this point the filePath and fileFormat will be changed. Don't use the delete method or we'll delete the new file.
                roiFile.delete()




import logging
import re
import typing as t_

from PyQt5 import QtCore
from PyQt5.QtCore import QThread
from PyQt5.QtGui import QFont
from PyQt5.QtWidgets import QDialog, QWidget, QPushButton, QLabel, QGridLayout, QLineEdit, QApplication, QFormLayout, \
    QHBoxLayout, QMessageBox

from pwspy_gui.PWSAnalysisApp.componentInterfaces import ROIManager
from pwspy_gui.sharedWidgets.dialogs import BusyDialog
from skimage import filters, morphology, measure
import matplotlib.pyplot as plt
import pwspy.dataTypes as pwsdt
import numpy as np

DESCRIPTION = \
    """
    This action will draw an ROI named (`bg`, 0) on all 
    selected cells which have a PWS analysis matching the 
    name you provide.
    
    Warning: Any existing ROIs saved as ("bg", 0) will be overwritten.
    """


class BGRoiDrawer:
    """
    Opens a dialog box to get information to begin automatic ROI drawing and then runs the process in a separate thread. A message
    box appears at the end giving some useful information.

    Args:
         parent: The QWidget to serve as the `parent` to the dialog boxes.
         roiManager: An object that manages saving and loading of rois within a context. If None is supplied then the ROI operations will be performed directly
    """
    def __init__(self, parent: QWidget = None, roiManager: ROIManager = None):
        self._parent = parent
        self._roiManager = roiManager
        self._processingThread = QThread(self._parent)

    def run(self, acqs: t_.Sequence[pwsdt.Acquisition]):
        analysisname = self._showDialog()
        if analysisname is None:
            return
        else:
            self._threadException = None
            def runInThread(acqList=acqs, anName=analysisname):
                try:
                    self.drawBackgroundROIs(acqList, anName)
                except Exception as e:
                    logging.getLogger(__name__).exception(e)
                    self._threadException = e

            def onFinished():
                if self._threadException is not None:
                    QMessageBox.information(self._parent, "Error!", str(self._threadException))
                else:
                    QMessageBox.information(self._parent, "Finished!", "Automatic background detection is completed.")

            self._processingThread.run = runInThread
            self._processingThread.finished.connect(onFinished)
            self._processingThread.start()
            QMessageBox.information(self._parent, "Be patient", 'A notification will appear when automatic background detection is complete.')

    def _showDialog(self) -> str:
        dlg = BGROIDialog(self._parent)
        accepted = dlg.exec() == QDialog.Accepted
        if accepted:
            return dlg.getAnalysisName()
        else:
            return None

    def drawBackgroundROIs(self, acqs: t_.Iterable[pwsdt.Acquisition], analysisNamePattern: str):
        """

        Args:
            acqs: A list of Acquisition object to run.
            analysisNamePattern: A regex pattern of the analysis file to use for detecting background
        """
        logger = logging.getLogger(__name__)
        skipped = 0
        for i, acq in enumerate(acqs):
            try:
                anNames = acq.pws.getAnalyses()
                anName = [name for name in anNames if re.match(analysisNamePattern, name)][0]  # Select the first analysis name that matches the regex pattern
                rms = acq.pws.loadAnalysis(anName).rms
            except IndexError:
                logger.info(f"Skipping {acq.filePath}. No PWS analysis matching {analysisNamePattern} found.")
                skipped += 1
                continue
            logger.info(f"Auto-Drawing background ROI for acquisition {i+1} of {len(acqs)}")
            thresh = filters.threshold_otsu(rms)
            bgMask = rms < thresh  # Our background should be everything below the threshold.
            bgMask = morphology.binary_erosion(bgMask, morphology.disk(15))
            smthDisk = morphology.disk(15)
            bgMask = morphology.binary_closing(morphology.binary_opening(bgMask, smthDisk), smthDisk)  # Smoothing
            label = measure.label(bgMask)  # Identify and label connected components
            try:
                largestLabel = np.argmax(np.bincount(label.flat)[1:]) + 1  # The number of the largest labeled region (We exclude the 0th item and then add 1 to avoid including the negative region (labeled 0))
            except ValueError: # If there are no labeled regions then we will get `attempt to get argmax of an empty sequence here.                logger.info(f"Skipping {acq.filePath}. No PWS analysis matching {analysisNamePattern} found.")
                logger.info(f"Skipping {acq.filePath}. No background region found.")
                skipped += 1
                continue
            mask = label == largestLabel  # Our background region should be the largest non-zero region
            roi = pwsdt.Roi.fromMask(mask=mask)
            if self._roiManager is None:
                acq.saveRoi('bg', 0, roi, overwrite=True)
            else:
                self._roiManager.createRoi(acq, roi, 'bg', 0, True)
        logger.info(f"Skipped {skipped} of {len(acqs)} acquisitions")


class BGROIDialog(QDialog):
    def __init__(self, parent: QWidget = None):
        super().__init__(parent=parent)
        self.setWindowTitle("Automatic Background Detection")

        self.description = QLabel(DESCRIPTION)
        font = self.description.font()
        font.setBold(True)
        self.description.setFont(font)

        self.analysisName = QLineEdit()

        self.okButton = QPushButton("Ok")
        self.okButton.released.connect(self.accept)

        self.cancelButton = QPushButton("Cancel")
        self.cancelButton.released.connect(self.reject)

        grid = QGridLayout()
        grid.addWidget(self.description, 0, 0)
        form = QFormLayout()
        form.addRow("Analysis Name (regex):", self.analysisName)
        grid.addLayout(form, 1, 0)
        ll = QHBoxLayout()
        ll.addWidget(self.okButton)
        ll.addWidget(self.cancelButton)
        grid.addLayout(ll, 2, 0)
        self.setLayout(grid)

    def getAnalysisName(self) -> str:
        return self.analysisName.text()


if __name__ == '__main__':
    import sys
    app = QApplication(sys.argv)
    g = BGRoiDrawer()
    g.run()
    a = 1



# Copyright 2018-2020 Nick Anthony, Backman Biophotonics Lab, Northwestern University
#
# This file is part of PWSpy.
#
# PWSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PWSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PWSpy.  If not, see <https://www.gnu.org/licenses/>.

from __future__ import annotations

import logging
from datetime import datetime

from PyQt5 import QtGui, QtCore
from PyQt5.QtWidgets import QTableWidget, QAbstractItemView, QApplication, QTableWidgetItem

import pwspy
import numpy as np

class CopyableTable(QTableWidget):
    def __init__(self):
        super().__init__()
        self.setSelectionMode(QAbstractItemView.ContiguousSelection)

    def keyPressEvent(self, event):
        if event.matches(QtGui.QKeySequence.Copy):
            self.copy()
        else:
            super().keyPressEvent(event)

    def copy(self):
        try:
            sel = self.selectedRanges()[0]
            t = '\t'.join(
                [self.horizontalHeaderItem(i).text() for i in range(sel.leftColumn(), sel.rightColumn() + 1) if not self.isColumnHidden(i)]) + '\n'
            for i in range(sel.topRow(), sel.bottomRow() + 1):
                for j in range(sel.leftColumn(), sel.rightColumn() + 1):
                    if not self.isColumnHidden(j):
                        if t[-1] != '\n': t += '\t'
                        item = self.item(i, j)
                        t += ' ' if item is None else item.text()
                t += '\n'
            QApplication.clipboard().setText(t)
        except Exception as e:
            logger = logging.getLogger(__name__)
            logger.warning("Copy Failed: ", e)


class NumberTableWidgetItem(QTableWidgetItem):
    """This table widget item will be sorted numerically rather than alphabetically (1, 10, 11, 2, ...)"""
    def __init__(self, num: float = None):
        if num is None:
            super().__init__("None")
            num = -np.inf
        else:
            super().__init__(str(num))
            num = float(num)  # in case the constructor is called with a string.
        self.setFlags(QtCore.Qt.ItemIsSelectable | QtCore.Qt.ItemIsEnabled)  # read only
        self.number = num

    def __lt__(self, other: 'NumberTableWidgetItem'):
        return self.number < other.number

    def __gt__(self, other: 'NumberTableWidgetItem'):
        return self.number > other.number

    def setNumber(self, num: float):
        self.number = num
        self.setText(str(num))


class DatetimeTableWidgetItem(QTableWidgetItem):
    """This table widget item will be sorted chronologically rather than alphabetically."""
    def __init__(self, dtime: datetime):
        if isinstance(dtime, str):
            dtime = datetime.strptime(dtime, pwspy.dateTimeFormat) #If constructor called with a string convert to datetime.
        super().__init__(datetime.strftime(dtime, pwspy.dateTimeFormat))
        self.setFlags(QtCore.Qt.ItemIsSelectable | QtCore.Qt.ItemIsEnabled)  # read only
        self.dtime = dtime

    def __lt__(self, other: DatetimeTableWidgetItem):
        return self.dtime < other.dtime

    def __gt__(self, other: DatetimeTableWidgetItem):
        return self.dtime > other.dtime



# Copyright 2018-2020 Nick Anthony, Backman Biophotonics Lab, Northwestern University
#
# This file is part of PWSpy.
#
# PWSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PWSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PWSpy.  If not, see <https://www.gnu.org/licenses/>.

from __future__ import annotations

import json
import logging
import traceback
from typing import Optional

from PyQt5 import QtCore, QtGui
from PyQt5.QtCore import QModelIndex, QPoint
from PyQt5.QtWidgets import QDialog, QWidget, QHBoxLayout, QPushButton, QAbstractItemView, QTableView, QVBoxLayout, \
    QMenu, QAction, QMessageBox
import pandas as pd
import typing
import numpy as np

from pwspy_gui.sharedWidgets.extraReflectionManager.ERIndex import ERIndex
from pwspy_gui.sharedWidgets.extraReflectionManager._ERDataDirectory import ERDataDirectory
from pwspy_gui.sharedWidgets.extraReflectionManager.ERDataComparator import ERDataComparator
from pwspy.dataTypes import ExtraReflectanceCube
from mpl_qt_viz.visualizers import PlotNd


if typing.TYPE_CHECKING:
    from pwspy_gui.sharedWidgets.extraReflectionManager import ERManager


class ERUploaderWindow(QDialog):
    """This window provides the user a visual picture of the local and online status of the Extra Reflectance calibration file repository. It also allows uploading
    of files that are present locally but not on the server. It does not have good handling of edge cases, e.g. online server in inconsistent state."""
    def __init__(self, manager: ERManager, parent: Optional[QWidget] = None):
        self._dataComparator = ERDataComparator(manager.onlineDirectory, manager.localDirectory)
        self._manager = manager
        self._selectedId: str = None
        super().__init__(parent)
        self.setModal(False)
        self.setWindowTitle("Extra Reflectance File Manager")
        self.setLayout(QVBoxLayout())
        self.table = QTableView(self)
        self.fileStatus = self._dataComparator.compare()
        self.table.setModel(PandasModel(self.fileStatus))
        self.table.setSelectionMode(QTableView.SingleSelection)
        self.table.setSelectionBehavior(QTableView.SelectRows)
        self.table.customContextMenuRequested.connect(self.openContextMenu)
        self.table.setContextMenuPolicy(QtCore.Qt.CustomContextMenu)
        self.uploadButton = QPushButton("Upload to Drive")
        self.uploadButton.released.connect(self._updateGDrive)
        self.uploadButton.setToolTip("Upload any files for which the status is `Local Only` to the google drive account.")
        self.refreshButton = QPushButton('Refresh')
        self.refreshButton.setToolTip("Rescan the files in the applications data directory.")
        self.refreshButton.released.connect(self.refresh)
        self.updateIndexButton = QPushButton("Update Index File")
        self.updateIndexButton.setToolTip("Download the online index file and merge it with our local index file.")
        self.updateIndexButton.released.connect(self._updateIndexFile)
        self.layout().addWidget(self.table)
        l = QHBoxLayout()
        l.setContentsMargins(0, 0, 0, 0)
        l.addWidget(self.uploadButton)
        l.addWidget(self.updateIndexButton)
        l.addWidget(self.refreshButton)
        w = QWidget()
        w.setLayout(l)
        self.layout().addWidget(w)
        self.table.setColumnWidth(0, 300); self.table.setColumnWidth(1, 150); self.table.setColumnWidth(2, 150)
        self.table.setMinimumWidth(sum(self.table.columnWidth(i)for i in range(self.table.model().columnCount())) + self.table.verticalHeader().width() + 20)

    def displayInfo(self, index: QModelIndex):
        msg = QMessageBox.information(self, 'Info', repr(self.fileStatus.iloc[index.row()]))

    def plotData(self, index: QModelIndex):
        idTag = self.fileStatus.iloc[index.row()]['idTag']
        md = self._dataComparator.local.getMetadataFromId(idTag)
        erCube = ExtraReflectanceCube.fromMetadata(md)
        self.plotHandle = PlotNd(erCube.data)

    def openContextMenu(self, pos: QPoint):
        """This method opens a context menu, it should be called when the user right clicks."""
        index = self.table.indexAt(pos)
        row = self.fileStatus.iloc[index.row()]
        menu = QMenu()
        displayAction = QAction("Display Info")
        displayAction.triggered.connect(lambda: self.displayInfo(index))
        menu.addAction(displayAction)
        if row['Local Status'] == self._dataComparator.local.DataStatus.found.value:
            plotAction = QAction("Plot Local Data")
            plotAction.triggered.connect(lambda: self.plotData(index))
            menu.addAction(plotAction)
        menu.exec(self.mapToGlobal(pos))

    def _updateGDrive(self):
        """Checks for all files taht are present locally but not on the server. Uploads those file and then overwrites the index."""
        try:
            status = self.fileStatus
            if not np.all((status['Index Comparison'] == ERDataComparator.ComparisonStatus.LocalOnly.value) | (status['Index Comparison'] == ERDataComparator.ComparisonStatus.Match.value)):
                raise ValueError("Uploading cannot be performed if the local index file is not valid. Try updating the index file.")
            uploadableRows = (status['Index Comparison'] == ERDataComparator.ComparisonStatus.LocalOnly.value) | (status['Online Status'] == ERDataDirectory.DataStatus.missing.value)
            if np.any(uploadableRows):  # There is something to upload
                for i, row, in status.loc[uploadableRows].iterrows():
                    fileName = [i.fileName for i in self._dataComparator.local.index.cubes if i.idTag == row['idTag']][0]
                    self._manager.upload(fileName)
                self._manager.upload('index.json')
            self.refresh()
        except Exception as e:
            logger = logging.getLogger(__name__)
            logger.exception(e)
            mess = QMessageBox.information(self, 'Sorry', str(e))

    def refresh(self):
        """Scans local and online files to refresh the display."""
        self._dataComparator.updateIndexes()
        self.fileStatus = self._dataComparator.compare()
        self.table.setModel(PandasModel(self.fileStatus))

    def _updateIndexFile(self):
        self._dataComparator.online.updateIndex()
        index = ERIndex.merge(self._dataComparator.local.index, self._dataComparator.online.index)
        self._dataComparator.local.saveNewIndex(index)
        self.refresh()


class PandasModel(QtCore.QAbstractTableModel):
    """This Qt model is used to adapt a Pandas dataframe to a QTableView widget. This was found somewhere on stack overflow."""
    def __init__(self, df=pd.DataFrame(), parent=None):
        """`df` should be the dataframe that you want to view."""
        QtCore.QAbstractTableModel.__init__(self, parent=parent)
        self._df = df

    def headerData(self, section, orientation, role=QtCore.Qt.DisplayRole):
        if role != QtCore.Qt.DisplayRole:
            return QtCore.QVariant()

        if orientation == QtCore.Qt.Horizontal:
            try:
                return self._df.columns.tolist()[section]
            except (IndexError, ):
                return QtCore.QVariant()
        elif orientation == QtCore.Qt.Vertical:
            try:
                # return self.df.index.tolist()
                return self._df.index.tolist()[section]
            except (IndexError, ):
                return QtCore.QVariant()

    def _calculateColor(self, index: QModelIndex):
        if self._df.columns[index.column()] == 'idTag':
            c = QtGui.QColor('white')
        elif (self._df.iloc[index.row(), index.column()] == ERDataDirectory.DataStatus.found.value) or (self._df.iloc[index.row(), index.column()] == ERDataComparator.ComparisonStatus.Match.value):
            c = QtGui.QColor('green')
        else:
            c = QtGui.QColor('red')
        return c

    def data(self, index, role=QtCore.Qt.DisplayRole):
        if role == QtCore.Qt.BackgroundRole:
            color = self._calculateColor(index)
            return QtGui.QBrush(color)
        elif role != QtCore.Qt.DisplayRole:
            return QtCore.QVariant()

        if not index.isValid():
            return QtCore.QVariant()

        return QtCore.QVariant(str(self._df.iloc[index.row(), index.column()]))

    def setData(self, index, value, role):
        row = self._df.index[index.row()]
        col = self._df.columns[index.column()]
        if hasattr(value, 'toPyObject'):
            # PyQt4 gets a QVariant
            value = value.toPyObject()
        else:
            # PySide gets an unicode
            dtype = self._df[col].dtype
            if dtype != object:
                value = None if value == '' else dtype.type(value)
        self._df.set_value(row, col, value)
        return True

    def rowCount(self, parent=QtCore.QModelIndex()):
        return len(self._df.index)

    def columnCount(self, parent=QtCore.QModelIndex()):
        return len(self._df.columns)

    def sort(self, column, order):
        colname = self._df.columns.tolist()[column]
        self.layoutAboutToBeChanged.emit()
        self._df.sort_values(colname, ascending= order == QtCore.Qt.AscendingOrder, inplace=True)
        self._df.reset_index(inplace=True, drop=True)
        self.layoutChanged.emit()


# Copyright © 2018-2020 Nick Anthony, Backman Biophotonics Lab, Northwestern University
#
# This file is part of PWSpy.
#
# PWSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PWSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PWSpy.  If not, see <https://www.gnu.org/licenses/>.

"""
This module contains Abstract classes used to define the methods that should be implemented by various objects that are central to the PWSAnalysisApp.

@author: Nick Anthony
"""
from __future__ import annotations
import abc
import typing
from typing import List, Optional
import sip
from PyQt5.QtCore import pyqtSignal
from pwspy import dataTypes as pwsdt
from pwspy.dataTypes import RoiFile
from pwspy_gui.PWSAnalysisApp.utilities.conglomeratedAnalysis import ConglomerateCompilerResults, \
    ConglomerateCompilerSettings
if typing.TYPE_CHECKING:
    from pwspy_gui.PWSAnalysisApp._dockWidgets.AnalysisSettingsDock import AbstractRuntimeAnalysisSettings


class QABCMeta(sip.wrappertype, abc.ABCMeta):
    """This metaclass allows us to use abstract classes alongside Qt"""
    pass


class CellSelector(metaclass=QABCMeta):
    @abc.abstractmethod
    def loadNewCells(self, fileNames: List[str], workingDir: str): pass

    @abc.abstractmethod
    def getSelectedCellMetas(self) -> List[pwsdt.Acquisition]: pass

    @abc.abstractmethod
    def getAllCellMetas(self) -> List[pwsdt.Acquisition]: pass

    @abc.abstractmethod
    def getSelectedReferenceMeta(self) -> Optional[pwsdt.Acquisition]: pass

    @abc.abstractmethod
    def setSelectedCells(self, cells: List[pwsdt.Acquisition]): pass

    @abc.abstractmethod
    def setSelectedReference(self, ref: pwsdt.Acquisition): pass

    @abc.abstractmethod
    def setHighlightedCells(self, cells: List[pwsdt.Acquisition]): pass

    @abc.abstractmethod
    def setHighlightedReference(self, ref: pwsdt.Acquisition): pass

    @abc.abstractmethod
    def refreshCellItems(self, cells: List[pwsdt.Acquisition] = None):
        """`Cells` indicates which cells need refreshing. If cells is None then all cells will be refreshed."""
        pass

    @abc.abstractmethod
    def close(self): pass

    @abc.abstractmethod
    def getRoiManager(self) -> ROIManager:
        """Return the ROI manager that manages the saving and loading of ROIs"""
        pass


class ResultsTableController(metaclass=QABCMeta):
    @abc.abstractmethod
    def addCompilationResult(self, result: ConglomerateCompilerResults, acquisition: pwsdt.Acquisition): pass

    @abc.abstractmethod
    def clearCompilationResults(self): pass

    @abc.abstractmethod
    def getSettings(self) -> ConglomerateCompilerSettings: pass

    @abc.abstractmethod
    def getRoiName(self) -> str: pass

    @abc.abstractmethod
    def getAnalysisName(self) -> str: pass


class AnalysisSettingsCreator(metaclass=QABCMeta):
    @abc.abstractmethod
    def getListedAnalyses(self) -> typing.List[AbstractRuntimeAnalysisSettings]: pass


class ROIManager(metaclass=QABCMeta):
    """Handles the actual file saving and retrieval. Any code using this should only modify ROI files through this manager."""

    # Implementation should fire these events afer the deed is done.
    roiRemoved = pyqtSignal(RoiFile)
    roiUpdated = pyqtSignal(RoiFile)
    roiCreated = pyqtSignal(RoiFile, bool)  # bool: May have been an overwrite

    @abc.abstractmethod
    def removeRoi(self, roiFile: pwsdt.RoiFile):
        pass

    @abc.abstractmethod
    def updateRoi(self, roiFile: pwsdt.RoiFile, roi: pwsdt.Roi):
        pass

    @abc.abstractmethod
    def createRoi(self, acq: pwsdt.Acquisition, roi: pwsdt.Roi, roiName: str, roiNumber: int, overwrite: bool = False) -> pwsdt.RoiFile:
        pass

    @abc.abstractmethod
    def getROI(self, acq: pwsdt.Acquisition, roiName: str, roiNum: int) -> pwsdt.RoiFile:
        pass

    @abc.abstractmethod
    def close(self):
        """Make sure all files are wrapped up"""
        pass


import os
import traceback
from datetime import datetime

from PyQt5 import QtCore
from PyQt5.QtWidgets import QApplication, QFileDialog, QMessageBox, QListWidgetItem
from matplotlib import pyplot as plt

from pwspy_gui import appPath
from pwspy_gui.ExtraReflectanceCreator.ERWorkFlow import ERWorkFlow
from pwspy_gui.ExtraReflectanceCreator.widgets.mainWindow import MainWindow
from pwspy_gui.sharedWidgets.extraReflectionManager import ERManager


class ERApp(QApplication):
    """
    An application for generating the "ExtraReflectance" calibration files from a set of reference reflectance measurements.
    """
    def __init__(self, args):
        QApplication.setAttribute(QtCore.Qt.AA_EnableHighDpiScaling)
        super().__init__(args)
        plt.interactive(True)
        settings = QtCore.QSettings("BackmanLab", "ERCreator")
        try:
            initialDir = settings.value('workingDirectory')
        except TypeError: #Setting not found
            initialDir = None
        wDir = QFileDialog.getExistingDirectory(caption='Select the root `ExtraReflection` directory', directory=initialDir)
        settings.setValue("workingDirectory", wDir)
        self.checkDataDir()

        try:
            self.workflow = ERWorkFlow(wDir, self.gDriveDir)
            self.erManager = ERManager(self.gDriveDir)
            self.window = MainWindow(self.erManager)
            self.connectWindowToWorkflow()
        except Exception as e:
            QMessageBox.warning(None, "Error", repr(e))
            raise e

    def connectWindowToWorkflow(self):
        for k, v in self.workflow.fileStruct.items():
            self.window.listWidg.addItem(k)
        self.window.listWidg.currentItemChanged.connect(self.selectionChanged)
        self.window.deleteFigsButton.released.connect(self.workflow.deleteFigures)
        self.window.saveButton.released.connect(
            self._cb(
                lambda: self.workflow.save(
                    self.window.checkedSettings,
                    self.window.binning,
                    self.window.parallelProcessing,
                    self.window.numericalAperture.value(),
                    self.window)))
        self.window.selListWidg.itemChanged.connect(self.workflow.invalidateCubes)
        self.window.binningCombo.currentIndexChanged.connect(self.workflow.invalidateCubes)
        self.window.compareDatesButton.released.connect(
            self._cb(
                lambda: self.workflow.compareDates(
                    self.window.checkedSettings,
                    self.window.binning,
                    self.window.parallelProcessing)))
        self.window.plotButton.released.connect(
            self._cb(
                lambda: self.workflow.plot(
                    self.window.checkedSettings,
                    self.window.binning,
                    self.window.parallelProcessing,
                    self.window.numericalAperture.value())))

    def checkDataDir(self):
        self.homeDir = os.path.join(appPath, 'ExtraReflectanceCreatorData')
        if not os.path.exists(self.homeDir):
            os.mkdir(self.homeDir)
        self.gDriveDir = os.path.join(self.homeDir, 'GoogleDriveData')
        if not os.path.exists(self.gDriveDir):
            os.mkdir(self.gDriveDir)

    def _cb(self, func):
        """Return a wrapped function with extra gui stuff."""
        def newfunc():
            """Toggle button enabled state. load new data if selection has changed. run the callback."""
            try:
                self.window.setEnabled(False)
                func()
            except:
                traceback.print_exc()
                msg = QMessageBox.warning(self.window, "Don't panic", "An error occurred. Please see the console for details.")
            finally:
                self.window.setEnabled(True)
        return newfunc

    def selectionChanged(self, item: QListWidgetItem, oldItem: QListWidgetItem):
        settings = self.workflow.directoryChanged(item.text())
        self.window.selListWidg.clear()

        _, settings = zip(*sorted(zip([datetime.strptime(sett, "%m_%d_%Y") for sett in settings], settings)))
        for sett in settings:
            _ = QListWidgetItem(sett)
            _.setFlags(_.flags() | QtCore.Qt.ItemIsUserCheckable)
            _.setCheckState(QtCore.Qt.Unchecked)
            self.window.selListWidg.addItem(_)


import typing as t_

import numpy as np
from PyQt5.QtCore import QObject
from pwspy.utility.acquisition import SequencerStep, SequenceAcquisition, PositionsStep, TimeStep, SequencerCoordinate, IterableSequencerStep
import pwspy.dataTypes as pwsdt
from pwspy_gui.PWSAnalysisApp._roiManager import ROIManager


class Options(t_.NamedTuple):
    copyAlong: bool
    trackMovement: bool


class SequenceController:
    """A utility class to help with selected acquisitions from a sequence that includes a multiple position and time series. both are optional"""
    def __init__(self, sequence: SequencerStep, acqs: t_.Sequence[SequenceAcquisition]):
        self.sequence = sequence
        self.coordMap: t_.Dict[pwsdt.Acquisition, SequencerCoordinate] = {acq.acquisition: acq.sequencerCoordinate for acq in acqs}  # A dictionary of the sequence coords keyed by tha acquisition

        self.iterSteps: t_.Tuple[IterableSequencerStep] = tuple((step for step in sequence.iterateChildren() if isinstance(step, IterableSequencerStep)))
        self._indices = tuple(0 for step in self.iterSteps)

    def getIterationNames(self, idx: int) -> t_.Sequence[str]:
        return tuple((self.iterSteps[idx].getIterationName(i) for i in range(self.iterSteps[idx].stepIterations())))

    def setCoordinates(self, *idxs: t_.Sequence[int]) -> pwsdt.Acquisition:
        acq = self.getAcquisitionForIndices(*idxs)
        self._indices = idxs
        return acq

    def getIndicesForAcquisition(self, acq: t_.Union[SequenceAcquisition, pwsdt.Acquisition]) -> t_.Tuple[int, ...]:
        """Returns the iteration indices of the given acquisition in the form (timeIdx, posIdx)"""
        coord: SequencerCoordinate = acq.sequencerCoordinate if isinstance(acq, SequenceAcquisition) else self.coordMap[acq]
        idxs = []
        for step in self.iterSteps:
            idxs.append(coord.getStepIteration(step))
        return tuple(idxs)

    def getAcquisitionForIndices(self, *idxs: int) -> pwsdt.Acquisition:
        step: SequencerStep = self.iterSteps[np.argmax([len(i.getTreePath()) if i is not None else 0 for i in self.iterSteps])]  # The step that is furthest down the tree path
        coordRange = step.getCoordinate()
        for idx, step in zip(idxs, self.iterSteps):
            coordRange.setAcceptedIterations(step.id, [idx])
        for acq, coord in self.coordMap.items():
            if coord in coordRange:
                return acq
        raise ValueError(f"No acquisition was found to match indices: {idxs}")  # If we got this far then no matching acquisition was found.

    def getCurrentIndices(self) -> t_.Tuple[int, ...]:
        return self._indices


class RoiController(QObject):
    """Handles applying ROI changes across axes.

    """
    def __init__(self, seqController: SequenceController, initialOptions: Options, roiManager: ROIManager, parent: QObject = None):
        super().__init__(parent=parent)
        self._seqController = seqController
        self._options = list(initialOptions for a in seqController.iterSteps)  # One open for each axis.
        self._roiManager = roiManager

    def setOptions(self, axis: int, options: Options):
        self._options[axis] = options

    def getOptions(self, axis: int) -> Options:
        return self._options[axis]

    def setRoiChanged(self, acq: pwsdt.Acquisition, roiFile: pwsdt.RoiFile, overwrite: bool):
        idxs = self._seqController.getIndicesForAcquisition(acq)
        for i, (step, option, idx) in enumerate(zip(self._seqController.iterSteps, self._options, idxs)):
            if not option.copyAlong:
                continue
            for ii in range(idx+1, step.stepIterations()):
                newIdxs = idxs[:i] + (ii,) + idxs[i+1:]
                acq = self._seqController.getAcquisitionForIndices(*newIdxs)
                self._roiManager.createRoi(acq, roiFile.getRoi(), roiFile.name, roiFile.number, overwrite=overwrite)

    def deleteRoi(self, acq: pwsdt.Acquisition, roiFile: pwsdt.RoiFile):
        idxs = self._seqController.getIndicesForAcquisition(acq)

        for i, (step, option, idx) in enumerate(zip(self._seqController.iterSteps, self._options, idxs)):
            if not option.copyAlong:
                continue
            for ii in range(idx+1, step.stepIterations()):
                newIdxs = idxs[:i] + (ii,) + idxs[i+1:]
                acq = self._seqController.getAcquisitionForIndices(*newIdxs)
                roiSpecs = [(roiName, roiNum) for roiName, roiNum, fformat in acq.getRois()]
                if (roiFile.name, roiFile.number) in roiSpecs:
                    self._roiManager.removeRoi(self._roiManager.getROI(acq, roiFile.name, roiFile.number))



# Copyright 2018-2020 Nick Anthony, Backman Biophotonics Lab, Northwestern University
#
# This file is part of PWSpy.
#
# PWSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PWSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PWSpy.  If not, see <https://www.gnu.org/licenses/>.

from __future__ import annotations

import logging
from datetime import datetime
from typing import List, Optional

from PyQt5 import QtCore
from PyQt5.QtCore import QPoint
from PyQt5.QtWidgets import (QDialog, QMessageBox, QWidget, QVBoxLayout, QPushButton, QMenu, QAction, QTreeWidget,
                             QTreeWidgetItem, QFileDialog)
import pwspy
import pwspy.dataTypes as pwsdt
from .exceptions import OfflineError

import typing

from mpl_qt_viz.visualizers import PlotNd

if typing.TYPE_CHECKING:
    from pwspy_gui.sharedWidgets.extraReflectionManager import ERManager
    from pwspy_gui.sharedWidgets.extraReflectionManager.ERIndex import ERIndexCube


class ERTreeWidgetItem(QTreeWidgetItem):
    def __init__(self, fileName: str, description: str, idTag: str, name: str, downloaded: bool):
        super().__init__()
        self.fileName = fileName
        self.description = description
        self.idTag = idTag
        self.systemName = self.idTag.split('_')[1]  # We used to categorize the calibrations by system name but this is confusing because a single system can have multiple configurations
        self.datetime = datetime.strptime(self.idTag.split('_')[2], pwspy.dateTimeFormat)
        self.name = name
        configurationName = name.split("-")[:-1]  # We now categorize by configurationName, however this isn't explicitly saved in the index so we extract it from the file name. Of course, this breaks if anyone puts a "-" in the configuration name. It would be better to explicitly save it from the ERCreator app
        self.configurationName = configurationName[0] if len(configurationName) == 1 else "-".join(configurationName)

        self.setText(0, datetime.strftime(self.datetime, '%B %d, %Y'))
        self.setToolTip(0, '\n'.join([f'File Name: {self.fileName}', f'ID: {self.idTag}', f'Description: {self.description}']))

        if downloaded:
            #Item can be selected. Checkbox no longer usable
            self.setFlags(QtCore.Qt.ItemIsSelectable | QtCore.Qt.ItemIsEnabled)
            self.setCheckState(0, QtCore.Qt.PartiallyChecked)
        else:
            #Checkbox can be checked to allow downloading. Nothing else can be done.
            self.setFlags(QtCore.Qt.ItemIsUserCheckable | QtCore.Qt.ItemIsEnabled)
            self.setCheckState(0, QtCore.Qt.Unchecked)
        self._downloaded = downloaded

    @property
    def downloaded(self): return self._downloaded

    def __lt__(self, other: ERTreeWidgetItem):  # Needed for sorting by date
        return self.datetime < other.datetime

    def isChecked(self) -> bool:
        return self.checkState(0) == QtCore.Qt.Checked


class ERSelectorWindow(QDialog):
    selectionChanged = QtCore.pyqtSignal(object) #Usually an ERMetaData object, sometimes None

    # noinspection PyUnresolvedReferences
    def __init__(self, manager: ERManager, parent: Optional[QWidget] = None):
        self._manager = manager
        self._selectedMetadata: Optional[pwsdt.ERMetaData] = None
        super().__init__(parent)
        self.setModal(False)
        self.setWindowTitle("Extra Reflectance Selector")
        self.setLayout(QVBoxLayout())
        self.tree = QTreeWidget(self)
        self.tree.setHeaderHidden(True)
        self.tree.setContextMenuPolicy(QtCore.Qt.CustomContextMenu)
        self.tree.customContextMenuRequested.connect(self.showContextMenu)
        self.tree.itemSelectionChanged.connect(self._setAcceptButtonEnabled)
        self.downloadButton = QPushButton("Download Checked Items")
        self.downloadButton.released.connect(self._downloadCheckedItems)
        self.selectLocalButton = QPushButton("Select From File")
        self.selectLocalButton.released.connect(self._selectLocalFile)
        self.acceptSelectionButton = QPushButton("Accept Selection")
        self.acceptSelectionButton.released.connect(self.acceptCurrentSelection)
        self.acceptSelectionButton.setEnabled(False)  # This will become enabled once a valid button is selected.
        self.layout().addWidget(self.tree)
        self.layout().addWidget(self.downloadButton)
        self.layout().addWidget(self.selectLocalButton)
        self.layout().addWidget(self.acceptSelectionButton)
        try:
            self._manager.download('index.json', parentWidget=self)
        except OfflineError:
            logger = logging.getLogger(__name__)
            logger.warning('Offline Mode: Could not update `Extra Reflectance` index file. Connection to Google Drive failed.')
        self._initialize()

    def _initialize(self):
        self._items: List[ERTreeWidgetItem] = []
        self.tree.clear()
        self._manager.localDirectory.updateIndex()
        self.fileStatus = self._manager.localDirectory.getFileStatus(skipMD5=True)  # Skipping the md5 hash check should speed things up here.
        for item in self._manager.localDirectory.index.cubes:
            self._addItem(item)
        # Sort items by date
        for item in [self.tree.invisibleRootItem().child(i) for i in range(self.tree.invisibleRootItem().childCount())]:
            item.sortChildren(0, QtCore.Qt.AscendingOrder)
        ig = QTreeWidgetItem()
        ig.setText(0, "Ignore")
        self.tree.invisibleRootItem().addChild(ig)

    def _addItem(self, item: ERIndexCube):
        treeItem = ERTreeWidgetItem(fileName=item.fileName, description=item.description, idTag=item.idTag, name=item.name,
                                    downloaded=self.fileStatus[self.fileStatus['idTag'] == item.idTag].iloc[0]['Local Status'] == self._manager.localDirectory.DataStatus.found.value)
        self._items.append(treeItem)
        _ = self.tree.invisibleRootItem()
        if treeItem.configurationName not in [_.child(i).text(0) for i in range(_.childCount())]:
            sysNameItem = QTreeWidgetItem(_, [treeItem.configurationName])
            sysNameItem.setFlags(QtCore.Qt.ItemIsEnabled)  # Don't allow selecting
            _.addChild(sysNameItem)
        parent = [i for i in [_.child(i) for i in range(_.childCount())] if i.text(0) == treeItem.configurationName][0]
        parent.addChild(treeItem)

    def showContextMenu(self, pos: QPoint):
        widgetItem: ERTreeWidgetItem = self.tree.itemAt(pos)
        if not isinstance(widgetItem, ERTreeWidgetItem):
            return  # Some treeItems exist which are not our custom ERTreeWidgetItem. Not menu for these items.
        menu = QMenu(self)
        displayAction = QAction("Display Info")
        displayAction.triggered.connect(lambda: self.displayInfo(widgetItem))
        menu.addAction(displayAction)
        if widgetItem.downloaded:
            plotAction = QAction("Plot Data")
            plotAction.triggered.connect(lambda checked, wItem=widgetItem: self._plot3dData(wItem))
            menu.addAction(plotAction)
        menu.exec(self.mapToGlobal(pos))

    def displayInfo(self, item: ERTreeWidgetItem):
        message = QMessageBox.information(self, item.name, '\n\n'.join([f'FileName: {item.fileName}',
                                                                      f'ID Tag: {item.idTag}',
                                                                      f'Description: {item.description}']))

    def _plot3dData(self, widgetItem):
        er = pwsdt.ExtraReflectanceCube.fromHdfFile(self._manager._directory, widgetItem.name)
        PlotNd(er.data, indices=[range(er.data.shape[0]), range(er.data.shape[1]), er.wavelengths])

    def _downloadCheckedItems(self):
        try :
            for item in self._items:
                if item.isChecked() and not item.downloaded:
                    # If it is checked then it should be downloaded
                    self._manager.download(item.fileName, parentWidget=self)
        except OfflineError as e:
            QMessageBox.information(self, "OfflineMode", "Sorry, for obvious reasons you can't download extra reflection calibration files when running in offline mode.")
        self._initialize()

    def _selectLocalFile(self):
        import os
        filePath, filterPattern = QFileDialog.getOpenFileName(self, "Select an ExtraReflectance file.", os.path.expanduser("~"), "*.h5")
        try:
            wDir, name = pwsdt.ERMetaData.directory2dirName(filePath)
            md = pwsdt.ERMetaData.fromHdfFile(wDir, name)
            self.setSelection(md)
            self.accept()
        except Exception as e:
            msg = QMessageBox.warning(self, "Error", str(e))

    def acceptCurrentSelection(self) -> None:
        items = self.tree.selectedItems()
        if len(items) == 0:
            self.setSelection(None)
            super().accept()
        elif self.tree.selectedItems()[0].text(0) == 'Ignore':
            self.setSelection(None)
            super().accept()
        else:
            try:
                md = self._manager.getMetadataFromId(self.tree.selectedItems()[0].idTag)
                self.setSelection(md)
                super().accept()
            except IndexError:  # Nothing was selected
                msg = QMessageBox.information(self, 'Uh oh!', 'No item was selected!')

    def getSelectedMetadata(self) -> Optional[pwsdt.ERMetaData]:
        return self._selectedMetadata

    def setSelection(self, md: pwsdt.ERMetaData):
        self._selectedMetadata = md
        self.selectionChanged.emit(md)

    def _setAcceptButtonEnabled(self):
        item = None
        items = self.tree.selectedItems()
        if len(items) > 0:
            item = items[0] # There should never be more than one item selected.
        #If a selectable item was selected then we'll have it as item here. Otherwise item will be None
        self.acceptSelectionButton.setEnabled(item is not None)




# Copyright 2018-2020 Nick Anthony, Backman Biophotonics Lab, Northwestern University
#
# This file is part of PWSpy.
#
# PWSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PWSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PWSpy.  If not, see <https://www.gnu.org/licenses/>.

# -*- coding: utf-8 -*-
"""
Created on Tue Nov 20 13:13:34 2018

@author: Nick Anthony

Based off of pwspy_gui.ExtraReflectanceCreator.examples.4waycalib.py"""

from pwspy.utility.fileIO import loadAndProcess
from pwspy.utility.reflection import reflectanceHelper, Material
import pwspy.utility.reflection.extraReflectance as er
from pwspy.dataTypes import Roi
from glob import glob
import matplotlib.pyplot as plt
import os
from matplotlib.backends.backend_pdf import PdfPages
import pandas as pd
import random


def processIm(im):
    im.correctCameraEffects()
    im.normalizeByExposure()
    print(im.metadata.filePath)
    print(im.metadata.exposure)
    im.filterDust(.75)
    return im


if __name__ == '__main__':
    __spec__ = None  # This is sometimes needed for multiprocessing to work on windows.
    plt.ion()

    rootDir = r'G:\Aya_NAstudy'
    produceRextraCube = False  # Do you want to actually save the file?
    plotResults = True  # Do you want to plot information about the calibration?

    materials = [('air', Material.Air), ('water', Material.Water)]  # Map folder names to a "pwspy.moduleConsts.Material" value
    settingsNADict = {0.5: ['matchedNAi_largeNAc', 'matchedNAi_smallNAc'],
                0.243: ['smallNAi_largeNAc', 'smallNAi_smallNAc']} #The folder names and NAi for each setting tested.

    for na in settingsNADict.keys():
        settings = settingsNADict[na]
        fig, axes = plt.subplots(ncols=2)
        axes[0].set_ylabel('n')
        axes[0].set_xlabel('nm')
        axes[0].set_title("Index of Refraction")
        axes[1].set_ylabel('reflectance')
        axes[1].set_xlabel('nm')
        axes[1].set_title(f"Glass Interface Reflectance. NA={na}")
        [axes[0].plot(reflectanceHelper.getRefractiveIndex(mat), label=matName) for matName, mat in materials]
        axes[0].legend()
        for matName, mat in materials:
            r = reflectanceHelper.getReflectance(mat, Material.Glass, NA=na)
            axes[1].plot(r.index, r, label=matName)
        axes[1].legend()

        fileFrame = pd.DataFrame([{'setting': setting, 'material': m[1], 'cube': cube} for setting in settings for m in materials for cube in glob(
            os.path.join(rootDir, setting, 'er', m[0],'Cell*'))])  # Convert the folder structure to a dataframe labeled with NA setting, Material, and the PwsCube objects.
        df = loadAndProcess(fileFrame, processIm, parallel=True)  # Returns a dataframe matching the form of `fileFrame` except the filepaths have been replaced with PwsCube objects (The filepaths have been loaded and processed using the `processIms` function.).

        theoryR = er.getTheoreticalReflectances(list(zip(*materials))[1], df['cube'][0].wavelengths, numericalAperture=na)
        matCombos = er.generateMaterialCombos(list(zip(*materials))[1])
        if plotResults:
            roi = random.choice(df['cube']).selectLassoROI()
            er.plotExtraReflection(df, theoryR, matCombos, na, roi, plotReflectionImages=False)
            with PdfPages(os.path.join(rootDir, "figs.pdf")) as pp:
                for i in plt.get_fignums():
                    f = plt.figure(i)
                    f.set_size_inches(9, 9)
                    pp.savefig(f)
        if produceRextraCube:
            for sett in set(df['setting']):
                allCombos = er.getAllCubeCombos(matCombos, df[df['setting'] == sett])
                erCube, rextras, plots = er.generateRExtraCubes(allCombos, theoryR, numericalAperture=na)
                erCube.toHdfFile(rootDir, f'rextra_{sett}')

    a = 1



# Copyright 2018-2020 Nick Anthony, Backman Biophotonics Lab, Northwestern University
#
# This file is part of PWSpy.
#
# PWSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PWSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PWSpy.  If not, see <https://www.gnu.org/licenses/>.

from __future__ import annotations
from abc import ABC, abstractmethod, ABCMeta

import typing
from typing import Optional

from PyQt5.QtCore import QObject

from pwspy_gui.PWSAnalysisApp._dockWidgets.AnalysisSettingsDock.runtimeSettings import AbstractRuntimeAnalysisSettings

if typing.TYPE_CHECKING:
    from pwspy.analysis import AbstractAnalysisSettings

class QtAbstractMeta(ABCMeta, type(QObject)):
    """Metaclass that allows implementing ABC and QObject simultaneously"""
    pass


class AbstractSettingsFrame(metaclass=QtAbstractMeta):
    @abstractmethod
    def loadFromSettings(self, settings: AbstractAnalysisSettings):
        """Populate the UI so that they match `settings`"""
        pass

    @abstractmethod
    def getSettings(self) -> AbstractRuntimeAnalysisSettings:
        """Generate a new settings object based on the current state of the UI"""
        pass



# Copyright 2018-2020 Nick Anthony, Backman Biophotonics Lab, Northwestern University
#
# This file is part of PWSpy.
#
# PWSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PWSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PWSpy.  If not, see <https://www.gnu.org/licenses/>.

from __future__ import annotations
from typing import List, Optional, Tuple
from PyQt5 import QtCore
from PyQt5.QtCore import QPoint
from PyQt5.QtWidgets import QListWidgetItem, QWidget, QScrollArea, QListWidget, QMessageBox, QMenu, QAction, QDialog, QGridLayout, QPushButton
from pwspy_gui.PWSAnalysisApp._dockWidgets.AnalysisSettingsDock.runtimeSettings import AbstractRuntimeAnalysisSettings, DynamicsRuntimeAnalysisSettings, PWSRuntimeAnalysisSettings
import typing
if typing.TYPE_CHECKING:
    from pwspy_gui.PWSAnalysisApp._dockWidgets import AnalysisSettingsDock


class AnalysisListItem(QListWidgetItem):
    def __init__(self, settings: AbstractRuntimeAnalysisSettings, parent: Optional[QWidget] = None):
        if isinstance(settings, PWSRuntimeAnalysisSettings):
            prefix = "PWS: "
        elif isinstance(settings, DynamicsRuntimeAnalysisSettings):
            prefix = "DYN: "
        else:
            raise TypeError(f"AnalysisListItem recieve unrecognized analyis settings object of type: {type(settings)}")
        super().__init__(prefix + settings.analysisName, parent)
        self.settings = settings


# noinspection PyUnresolvedReferences
class QueuedAnalysesFrame(QScrollArea):
    def __init__(self, parent: AnalysisSettingsDock):
        super().__init__()
        self.parent = parent
        self.listWidget = QListWidget()
        self.setWidget(self.listWidget)
        self.listWidget.setContextMenuPolicy(QtCore.Qt.CustomContextMenu)
        self.listWidget.customContextMenuRequested.connect(self.showContextMenu)
        self.listWidget.setMinimumHeight(30)
        self.setWidgetResizable(True)
        self.listWidget.itemDoubleClicked.connect(self.displayItemSettings)
        self.listWidget.itemClicked.connect(self.highlightAssociatedCells)

    @property
    def analyses(self) -> List[AbstractRuntimeAnalysisSettings]:
        return [self.listWidget.item(i).settings for i in range(self.listWidget.count())]

    def addAnalysis(self, settings: AbstractRuntimeAnalysisSettings):
        item = AnalysisListItem(settings, self.listWidget)  # the item is automatically added to the list here.

    def showContextMenu(self, point: QPoint):
        menu = QMenu("ContextMenu", self)
        deleteAction = QAction("Delete", self)
        deleteAction.triggered.connect(self.deleteSelected)
        menu.addAction(deleteAction)
        deleteAllAction = QAction("Delete All", self)
        deleteAllAction.triggered.connect(self.deleteAll)
        menu.addAction(deleteAllAction)
        menu.exec(self.mapToGlobal(point))

    def deleteSelected(self):
        for i in self.listWidget.selectedItems():
            self.listWidget.takeItem(self.listWidget.row(i))

    def deleteAll(self):
        self.listWidget.clear()

    def highlightAssociatedCells(self, item: AnalysisListItem):
        # Highlight relevant cells
        cellAcq = [cell.acquisitionDirectory for cell in item.settings.cellMetadata]
        refAcq = item.settings.referenceMetadata.acquisitionDirectory
        self.parent._selector.setHighlightedCells(cellAcq)
        self.parent._selector.setHighlightedReference(refAcq)

    def displayItemSettings(self, item: AnalysisListItem):
        # Open a dialog
        message = QMessageBox.information(self, item.settings.analysisName, item.settings.settings.toJsonString())
        # This used to show the settings in a dialog that could be edited. then some changes made it stop working and it was too much work to fix it.
        # d = QDialog(self.parent, (QtCore.Qt.WindowTitleHint | QtCore.Qt.Window))
        # d.setModal(True)
        # d.setWindowTitle(item.name)
        # l = QGridLayout()
        # if item.type == AnalysisTypes.PWS:
        #     settingsFrame = PWSSettingsFrame(self.parent.erManager)
        # elif item.type == AnalysisTypes.DYN:
        #     settingsFrame = DynamicsSettingsFrame(self.parent.erManager)
        # else:
        #     raise TypeError(f"Analysis of type {item.type} is not supported.")
        # settingsFrame.loadFromSettings(item.settings.getSaveableSettings())
        # settingsFrame._analysisNameEdit.setText(item.name)
        # settingsFrame._analysisNameEdit.setEnabled(False)  # Don't allow changing the name.
        #
        # okButton = QPushButton("OK")
        # okButton.released.connect(d.accept)
        #
        # l.addWidget(settingsFrame, 0, 0, 1, 1)
        # l.addWidget(okButton, 1, 0, 1, 1)
        # d.setLayout(l)
        # d.show()
        # d.exec()
        # try:
        #     item.settings = settingsFrame.getSettings()
        # except Exception as e:
        #     QMessageBox.warning(self.parent, 'Oh No!', str(e))






# Copyright 2018-2020 Nick Anthony, Backman Biophotonics Lab, Northwestern University
#
# This file is part of PWSpy.
#
# PWSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PWSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PWSpy.  If not, see <https://www.gnu.org/licenses/>.

from __future__ import annotations
import hashlib
import json
import logging
import os, io
from enum import Enum
from glob import glob
from typing import List

import pandas

from pwspy_gui.sharedWidgets.extraReflectionManager.ERIndex import ERIndex, ERIndexCube
from pwspy.dataTypes import ERMetaData
import typing
if typing.TYPE_CHECKING:
    from pwspy_gui.sharedWidgets.extraReflectionManager import ERDownloader
from abc import ABC, abstractmethod

class ERAbstractDirectory(ABC):
    """This class keeps track of the status of a directory that contains our extra reflection subtraction cubes.
    This can be implemented as a local directory on a hard drive or a folder on Google Drive, etc., whatever."""

    class DataStatus(Enum):
        md5Confict = 'Data MD5 mismatch'
        found = 'Found'
        notIndexed = 'Not Indexed'
        missing = 'Data File Missing'

    def __init__(self):
        self.index: ERIndex = None
        self.updateIndex()

    @abstractmethod
    def updateIndex(self):
        """update self.index from the `index.json` file."""
        pass

    @abstractmethod
    def getFileStatus(self) -> pandas.DataFrame:
        """return a dataframe indicating directory status by scanning the files in the directory being managed. Assumes that self.index is
        already up to date. In theory the `index.json` file has all the info we need, this method verifies that the
        `index.json` file is actually accurate."""
        pass


class ERDataDirectory(ERAbstractDirectory):
    """A class representing the locally stored data file directory for ExtraReflectanceCube files."""
    def __init__(self, directory: str):
        self._directory = directory
        super().__init__()

    def getMetadataFromId(self, idTag: str) -> ERMetaData:
        """Given the unique idTag string for an ExtraReflectanceCube this will search the index.json and return the
        ERMetaData file. If it cannot be found then an `IndexError will be raised."""
        try:
            match = [item for item in self.index.cubes if item.idTag == idTag][0]
        except IndexError:
            raise IndexError(f"An ExtraReflectanceCube with idTag {idTag} was not found in the index.json file at {self._directory}.")
        return ERMetaData.fromHdfFile(self._directory, match.name)

    def updateIndex(self):
        indexPath = os.path.join(self._directory, ERIndex.FILENAME)
        if not os.path.exists(indexPath):
            self.index = self._createNewIndexFile()
        else:
            try:
                with open(indexPath, 'r') as f:
                    self.index = ERIndex.load(f)
            except json.JSONDecodeError as e: # For unknown reasons the file sometime becomes corrupt. delete and start fresh
                logging.getLogger(__name__).warning("The ERIndex file was corrupt. Deleted and starting fresh.")
                os.remove(indexPath)
                self.index = self._createNewIndexFile()

    def _createNewIndexFile(self) -> ERIndex:
        index = ERIndex([])
        index.toJson(self._directory)
        return index

    def saveNewIndex(self, index: ERIndex):
        index.toJson(self._directory)
        self.index = index

    def getFileStatus(self, skipMD5: bool = False) -> pandas.DataFrame:
        files = glob(ERMetaData.dirName2Directory(self._directory, '*'))
        files = [(f, ERMetaData.validPath(f)) for f in files]  # validPath returns True/False in awhether the datacube was found.
        files = [(directory, name) for f, (valid, directory, name) in files if valid]  # Get rid of invalid files.
        files = [ERMetaData.fromHdfFile(directory, name) for directory, name in files]
        calculatedIndex = self._buildIndexFromFiles(files, skipMD5=skipMD5)
        d = self._compareIndexes(calculatedIndex, self.index, skipMD5=skipMD5)
        d = pandas.DataFrame(d).transpose()
        d.columns.values[1] = 'Local Status'
        return d

    @staticmethod
    def _buildIndexFromFiles(files: List[ERMetaData], skipMD5: bool = False) -> ERIndex:
        """Scan the data files in the directory and construct an ERIndex from the metadata. The `description` field is left blank though.
        Args:
            files (List[ERMetaData]): A list of the all the extra reflectance file objects that we want to construct an index from.
            skipMD5 (bool): If True then don't calculate the md5 hash for the files. This can be quite slow. Defaults to False.
        Returns:
            ERIndex: A new ERIndex representing the state of the extra reflectance files in `files`."""
        cubes = []
        for erCube in files:
            if skipMD5:
                md5 = None
            else:
                md5hash = hashlib.md5()
                with open(erCube.filePath, 'rb') as f:
                    md5hash.update(f.read())
                md5 = md5hash.hexdigest()
            cubes.append(ERIndexCube(erCube.filePath, erCube.inheritedMetadata['description'], erCube.idTag,
                                     erCube.directory2dirName(erCube.filePath)[-1], md5))
        return ERIndex(cubes)

    @staticmethod
    def _compareIndexes(ind1: ERIndex, ind2: ERIndex, skipMD5: bool = False) -> dict:
        """A utility function to compare two `ERIndex` objects and return a `dict` containing the status for each file.
        Args:
            ind1 (ERIndex): One of the ERIndexes to be compared.
            ind2 (ERIndex: The other ERIndex to be compared.
            skipMD5 (bool): If True then don't check if the MD5 hashes between the two indexes match. Sometimes we don't measure the MD5 index because it's slow. Defaults to False.
        Returns:
            pandas.Dataframe: A dataframe showing the results of the comparison for each item contained in at least on of the indexes."""
        foundTags = set([cube.idTag for cube in ind1.cubes])
        indTags = set([cube.idTag for cube in ind2.cubes])
        notIndexed = foundTags - indTags  # Tags in foundTags but not in indTags
        missing = indTags - foundTags  # Tags in in indTags but not in foundTags
        matched = indTags & foundTags  # Tags present in both sets
        dataMismatch = []  # Tags that match but have different md5 hashes
        for ID in matched:
            cube = [cube for cube in ind1.cubes if cube.idTag == ID][0]
            indCube = [cube for cube in ind2.cubes if cube.idTag == ID][0]
            if not skipMD5:
                if cube.md5 != indCube.md5:
                    dataMismatch.append(ID)
        # Construct a dataframe
        d = {}
        for i, tag, in enumerate(foundTags | indTags):
            if tag in missing:
                status = ERDataDirectory.DataStatus.missing.value
            elif tag in notIndexed:
                status = ERDataDirectory.DataStatus.notIndexed.value
            elif tag in dataMismatch:
                status = ERDataDirectory.DataStatus.md5Confict.value
            elif tag in matched:  # it must have been matched.
                status = ERDataDirectory.DataStatus.found.value
            else:
                raise Exception("Programming error.")  # This shouldn't be possible
            d[i] = {'idTag': tag, 'status': status}
        return d


class EROnlineDirectory(ERAbstractDirectory):
    """A class representing the status of the google drive directory"""
    def __init__(self, downloader: ERDownloader):
        self._downloader = downloader
        super().__init__()

    def updateIndex(self):
        with io.BytesIO() as f:
            f = self._downloader.downloadToRam('index.json', f)
            f.seek(0) #Move back to the beginning of the stream for reading.
            self.index = ERIndex.load(f)

    def getFileStatus(self) -> pandas.DataFrame:
        calculatedIndex = self._buildIndexFromOnlineFiles()
        d2 = self._compareIndexes(calculatedIndex, self.index)
        d2 = pandas.DataFrame(d2).transpose()
        d2.columns.values[1] = 'Online Status'
        return d2

    def _buildIndexFromOnlineFiles(self) -> ERIndex:
        """Return an ERIndex object from the HDF5 data files saved on Google Drive. No downloading required, just scanning metadata."""
        files = self._downloader.getFileMetadata()
        files = [f for f in files if ERMetaData.FILESUFFIX in f['name']]  # Select the dictionaries that correspond to a extra reflectance data file
        files = [ERIndexCube(fileName=f['name'], md5=f['md5Checksum'], name=ERMetaData.directory2dirName(f['name'])[-1], description=None, idTag=None) for f in files]
        return ERIndex(files)

    @staticmethod
    def _compareIndexes(calculatedIndex: ERIndex, jsonIndex: ERIndex) -> dict:
        """A utility function to compare two `ERIndex` objects and return a `dict` containing the status for each file.
        In this case we are not able to extract the idTags from the dataFiles without downloading them. use filenames instead"""
        foundNames = set([cube.fileName for cube in calculatedIndex.cubes])
        indNames = set([cube.fileName for cube in jsonIndex.cubes])
        notIndexed = foundNames - indNames  # fileNames in foundNames but not in indNames
        missing = indNames - foundNames  # fileNames in in indNames but not in foundNames
        matched = indNames & foundNames  # fileNames present in both sets
        dataMismatch = []  # fileNames that match but have different md5 hashes
        for name in matched:
            cube = [cube for cube in calculatedIndex.cubes if cube.fileName == name][0]
            indCube = [cube for cube in jsonIndex.cubes if cube.fileName == name][0]
            if cube.md5 != indCube.md5:
                dataMismatch.append(name)
        #  Construct a dataframe
        d = {}
        for i, fileName, in enumerate(foundNames | indNames):
            if fileName in missing:
                status = ERDataDirectory.DataStatus.missing.value
            elif fileName in notIndexed:
                status = ERDataDirectory.DataStatus.notIndexed.value
            elif fileName in dataMismatch:
                status = ERDataDirectory.DataStatus.md5Confict.value
            elif fileName in matched:  # it must have been matched.
                status = ERDataDirectory.DataStatus.found.value
            else:
                raise Exception("Programming error.")  # This shouldn't be possible
            d[i] = {'idTag': [ind.idTag for ind in jsonIndex.cubes if ind.fileName == fileName][0], 'status': status}

        return d



import typing as t_

from PyQt5 import QtCore
from PyQt5.QtCore import pyqtSignal
from PyQt5.QtWidgets import QFrame, QButtonGroup, QHBoxLayout, QPushButton, QScrollArea, QWidget, QSizePolicy


class ButtonBar(QFrame):
    buttonClicked = pyqtSignal(str, int)  # Indicates the iteration number and iteration name of the button that was clicked

    def __init__(self, items: t_.Sequence[str], parent=None):
        super().__init__(parent=parent)
        # self.setFrameStyle(QFrame.Panel)
        self._bGroup = QButtonGroup(self)
        l = QHBoxLayout(self)
        for i, itemName in enumerate(items):
            b = QPushButton(itemName, parent=self)
            b.setCheckable(True)  # Toggleable
            self._bGroup.addButton(b, id=i)
            l.addWidget(b)
        self._selectedButtonId = self._bGroup.id(self._bGroup.buttons()[0])
        self._bGroup.buttons()[0].click()  # Make sure at least one button is selected.
        self._bGroup.buttonClicked.connect(self._buttonSelected)
        l.setSpacing(1)  # Move buttons close together
        l.setContentsMargins(0, 0, 0, 0)
        w = QFrame(self)
        w.setFrameStyle(QFrame.Box)
        w.setLayout(l)
        scrollArea = QScrollArea(parent=self)
        scrollArea.setVerticalScrollBarPolicy(QtCore.Qt.ScrollBarAlwaysOff)
        scrollArea.setStyleSheet("""QScrollBar:horizontal {
             height:10px;     
         }""")
        scrollArea.setWidget(w)
        scrollArea.setFixedHeight(10 + w.height())
        ll = QHBoxLayout()
        ll.setContentsMargins(0, 0, 0, 0)
        ll.addWidget(scrollArea)
        self.setLayout(ll)


    def _buttonSelected(self, btn):
        self._selectedButtonId = self._bGroup.id(btn)
        self.buttonClicked.emit(btn.text(), self._bGroup.id(btn))

    def getSelectedButtonId(self):
        return self._selectedButtonId

    def selectNextButton(self):
        id = self._selectedButtonId + 1
        if id >= len(self._bGroup.buttons()):
            id = 0
        self.setButtonSelected(id)

    def setButtonSelected(self, ID: int):
        btn = self._bGroup.button(ID)
        btn.click()


# Copyright 2018-2020 Nick Anthony, Backman Biophotonics Lab, Northwestern University
#
# This file is part of PWSpy.
#
# PWSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PWSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PWSpy.  If not, see <https://www.gnu.org/licenses/>.

import logging
import os
from typing import List

from PyQt5 import QtCore
from PyQt5.QtWidgets import QDockWidget, QWidget, QHBoxLayout, QScrollArea, QVBoxLayout, QPushButton, QMessageBox, \
    QLabel, QLineEdit, QSizePolicy, QButtonGroup, QFrame, QApplication

import pwspy.dataTypes as pwsdt
from pwspy_gui.PWSAnalysisApp.sharedWidgets.plotting import RoiDrawer
from pwspy_gui.sharedWidgets.utilityWidgets import AspectRatioWidget
from .widgets.littlePlot import LittlePlot
from ...componentInterfaces import CellSelector
from ...utilities.conglomeratedAnalysis import ConglomerateAnalysisResults


# noinspection PyUnresolvedReferences
class PlottingDock(QDockWidget):
    def __init__(self, selector: CellSelector):
        super().__init__("Plotting")
        self.selector = selector
        self.setStyleSheet("QDockWidget > QWidget { border: 1px solid lightgray; }")
        self.roiDrawer = None
        self.setObjectName('PlottingWidget')
        self._plots = []
        self.cellMetas = []
        self._widget = QWidget()
        self._widget.setLayout(QHBoxLayout())
        self._plotScroll = QScrollArea()
        self._plotScroll.setVerticalScrollBarPolicy(QtCore.Qt.ScrollBarAlwaysOn)
        self._plotScroll.setWidgetResizable(True)
        self._plotScroll.horizontalScrollBar().setEnabled(False)
        self._plotScroll.horizontalScrollBar().setVisible(False)
        self._plotScroll.setMinimumWidth(75)
        self._plotScroll.setMaximumWidth(600)
        self._scrollContents = AspectRatioWidget(1, self)
        self._scrollContents.setLayout(QVBoxLayout())
        a = QSizePolicy()
        # a.setHeightForWidth(True)
        a.setHorizontalPolicy(QSizePolicy.Maximum) #Stretch out as much as possible horizontally
        self._scrollContents.setSizePolicy(a)

        self._plotScroll.setWidget(self._scrollContents)
        buttons = QWidget()
        buttons.setMaximumWidth(100)
        buttons.setLayout(QVBoxLayout())
        _ = buttons.layout().addWidget
        self._anNameEdit = QLineEdit(self)
        self._oldPattern = ''
        self._oldCells = None
        self._anNameEdit.editingFinished.connect(self._anNameEditFinished)
        self._buttonGroup = QButtonGroup()
        self._lastButton = None
        self._buttonGroup.buttonReleased.connect(self._handleButtons)
        self._plotThumbnailButton = QPushButton("Thumbnail")
        self._plotRMSButton = QPushButton("RMS")
        self._plotRButton = QPushButton('R')
        self._buttonGroup.addButton(self._plotThumbnailButton)
        self._buttonGroup.addButton(self._plotRMSButton)
        self._buttonGroup.addButton(self._plotRButton)
        [i.setCheckable(True) for i in self._buttonGroup.buttons()]
        self._buttonGroup.buttons()[0].setChecked(True)
        frame = QFrame(parent=self)
        frame.setFrameStyle(QFrame.StyledPanel | QFrame.Plain)
        l = QVBoxLayout()
        [l.addWidget(i) for i in self._buttonGroup.buttons()]
        frame.setLayout(l)
        self._refreshButton = QPushButton("Refresh")
        self._refreshButton.released.connect(lambda: self._generatePlots(self.selector.getSelectedCellMetas()))
        self._roiButton = QPushButton("Draw ROI's")
        self._roiButton.released.connect(self._startRoiDrawing)
        label = QLabel("Analysis Name")
        label.setMaximumHeight(20)
        _(label)
        _(self._anNameEdit)
        _(self._refreshButton)
        _(self._roiButton)
        _(frame)

        self._widget.layout().addWidget(self._plotScroll)
        self._widget.layout().addWidget(buttons)
        # self._widget.setMaximumWidth(self.plotScroll.maximumWidth()+buttons.maximumWidth()+10)
        self.setWidget(self._widget)

        self._enableAnalysisPlottingButtons('false')

    def _addPlots(self, plots: List[QWidget]):
        self._plots.extend(plots)
        [self._scrollContents.layout().addWidget(plot) for plot in plots]
        self._plotsChanged()

    def _plotsChanged(self):
        if len(self._plots) > 0:
            self._scrollContents.setAspect(1 / len(self._plots))

    def _startRoiDrawing(self):
        metadatas = [(p.acq, p.analysis) for p in self._plots]
        if len(metadatas) > 0:  # Otherwise we crash
            try:
                mainWindow = QApplication.instance().window
                self.roiDrawer = RoiDrawer(metadatas, QApplication.instance().roiManager, parent=mainWindow)
            except Exception as e:
                logging.getLogger(__name__).exception(e)
                QMessageBox.information(self, "Error", "An error occured. Please see the log file.")
        else:
            QMessageBox.information(self, "Oops", "Please select which cells to plot.")

    def _anNameEditFinished(self):
        #Sometimes the signal gets sent twice. only generate plots if the string has changed.
        pattern = self._anNameEdit.text()
        if pattern != self._oldPattern or self.cellMetas != self._oldCells:
            # Determine if a name has been entered
            # if self.anNameEdit.text().strip() == '':
            #     self.enableAnalysisPlottingButtons(False)
            # else:
            #     self.enableAnalysisPlottingButtons(True)
            self._generatePlots(self.selector.getSelectedCellMetas())
        self._oldPattern = pattern
        self._oldCells = self.cellMetas

    def _enableAnalysisPlottingButtons(self, enable: str):
        enable = enable.lower()
        if enable == 'false': #TODO what is the usage of this state
            for button in self._buttonGroup.buttons():
                button.setEnabled(False)
            self._roiButton.setEnabled(False)
        elif enable == 'partial':
            for button in self._buttonGroup.buttons():
                if not button is self._plotThumbnailButton:
                    button.setEnabled(False)
            self._plotThumbnailButton.setEnabled(True)
            self._roiButton.setEnabled(True)
        elif enable == 'true':
            for button in self._buttonGroup.buttons():
                if not button is self._plotThumbnailButton:
                    button.setEnabled(True)
            self._plotThumbnailButton.setEnabled(True)
            self._roiButton.setEnabled(True)
        else:
            raise ValueError("`enable` string not recognized.")

    def _generatePlots(self, cells: List[pwsdt.Acquisition]):
        try:
            self.cellMetas = cells
            analysisName = self._anNameEdit.text()
            #clear Plots
            for i in self._plots:
                self._scrollContents.layout().removeWidget(i)
                i.deleteLater()
            self._plots = []
            if len(self.cellMetas) == 0:
                messageBox = QMessageBox(self)
                messageBox.information(self, "Oops!", "Please select the cells you would like to plot.")
                messageBox.setFixedSize(500, 200)
            if len(cells) > 0:
                buttonState = 'partial'
            else:
                buttonState = 'false'
            plotsToAdd = []
            for cell in cells:
                if analysisName.strip() == '': #No analysis name was entered. don't load an analysis, just the thumbnail
                    plotsToAdd.append(LittlePlot(cell, ConglomerateAnalysisResults(None, None), f"{os.path.split(cell.filePath)[-1]}"))
                else:
                    dynAnalysis = pwsAnalysis = None
                    if cell.pws is not None:
                        if analysisName in cell.pws.getAnalyses():
                            pwsAnalysis = cell.pws.loadAnalysis(analysisName)
                    if cell.dynamics is not None:
                        if analysisName in cell.dynamics.getAnalyses():
                            dynAnalysis = cell.dynamics.loadAnalysis(analysisName)
                    analysis = ConglomerateAnalysisResults(pwsAnalysis, dynAnalysis)
                    if pwsAnalysis is None and dynAnalysis is None: #Specified analysis was not found, load a dummy widget
                        plotsToAdd.append(LittlePlot(cell, ConglomerateAnalysisResults(None, None), f"{analysisName} {os.path.split(cell.filePath)[-1]}", "Analysis Not Found!"))
                    else:
                        plotsToAdd.append(LittlePlot(cell, analysis, f"{analysisName} {os.path.split(cell.filePath)[-1]}"))
                        buttonState = 'true'  # Enable all buttons if there were some valid analysis.
            self._addPlots(plotsToAdd)
            self._enableAnalysisPlottingButtons(buttonState)
        except Exception as e:
            logger = logging.getLogger(__name__)
            logger.exception(e)
            QMessageBox.information(self, "Error!", str(e))

    def _handleButtons(self, button: QPushButton):
        if button != self._lastButton:
            for plot in self._plots:
                try:
                    if button is self._plotThumbnailButton:
                        plot.changeData(plot.PlotFields.Thumbnail)
                    elif button is self._plotRMSButton:
                        plot.changeData(plot.PlotFields.RMS)
                    elif button is self._plotRButton:
                        plot.changeData(plot.PlotFields.MeanReflectance)
                except ValueError:  # The analysis field wasn't found
                    plot.changeData(plot.PlotFields.Thumbnail)
            self._lastButton = button

    def setAnalysisName(self, name: str):
        self._anNameEdit.setText(name)

    def refreshPlots(self):
        self._generatePlots(self.selector.getSelectedCellMetas())


# Copyright 2018-2020 Nick Anthony, Backman Biophotonics Lab, Northwestern University
#
# This file is part of PWSpy.
#
# PWSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PWSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PWSpy.  If not, see <https://www.gnu.org/licenses/>.
from PyQt5 import QtGui
import os
from PyQt5.QtWidgets import QMainWindow, QWidget, QGridLayout, QListWidget, QComboBox, QPushButton, \
    QLabel, QDoubleSpinBox, QCheckBox

from pwspy_gui import resources
from pwspy_gui.sharedWidgets.extraReflectionManager import ERManager


class MainWindow(QMainWindow):
    """This is the main window of the ERCreator app.
    Args:
        manager (ERmanager): The class which handles organization of the Extra reflectance database locally an online."""
    def __init__(self, manager: ERManager):
        super().__init__()
        self.explorerWindow = manager.createManagerWindow(self)
        self.setWindowTitle("Extra Reflectance Creator")
        self.setWindowIcon(QtGui.QIcon(os.path.join(resources, 'cellLogo.png')))

        widg = QWidget()
        layout = QGridLayout()
        self.listWidg = QListWidget(self)
        self.selListWidg = QListWidget(self)
        self.binningCombo = QComboBox()
        self.binningCombo.addItems(['Auto', '1x1', '2x2', '3x3'])
        self.parallelCheckBox = QCheckBox("Parallel Process", self)
        self.parallelCheckBox.setToolTip("Uses significantly more ram but may be significantly faster")
        self.compareDatesButton = QPushButton("Compare Dates")
        self.plotButton = QPushButton("Plot Details")
        self.saveButton = QPushButton("Save Checked Dates")
        self.deleteFigsButton = QPushButton("Close Figures")
        self.viewFilesButton = QPushButton("View Files")
        self.viewFilesButton.released.connect(self.viewFiles)
        self.numericalAperture = QDoubleSpinBox()
        self.numericalAperture.setRange(0, 2)
        self.numericalAperture.setSingleStep(0.01)
        self.numericalAperture.setValue(0.52)
        row = 0
        layout.addWidget(self.listWidg, row, 0, 4, 4)
        layout.addWidget(self.selListWidg, row, 4, 4, 4)
        row += 4
        layout.addWidget(self.compareDatesButton, row, 0, 1, 1)
        layout.addWidget(self.plotButton, row, 1, 1, 1)
        layout.addWidget(self.deleteFigsButton, row, 2, 1, 1)
        layout.addWidget(QLabel("Binning"), row, 4, 1, 1)
        layout.addWidget(self.binningCombo, row, 5, 1, 1)
        layout.addWidget(self.parallelCheckBox, row, 6, 1, 1)
        row += 1
        layout.addWidget(self.saveButton, row, 0, 1, 1)
        layout.addWidget(self.viewFilesButton, row, 1, 1, 1)
        layout.addWidget(QLabel("NA"), row, 4, 1, 1)
        layout.addWidget(self.numericalAperture, row, 5, 1, 1)
        widg.setLayout(layout)
        self.setCentralWidget(widg)
        self.buttons = [self.compareDatesButton, self.plotButton, self.saveButton]
        self.show()

    @property
    def binning(self) -> int:
        num = self.binningCombo.currentIndex()
        return num if num != 0 else None

    @property
    def parallelProcessing(self) -> bool:
        return self.parallelCheckBox.isChecked()

    @property
    def checkedSettings(self):
        dateItems = [self.selListWidg.item(i) for i in range(self.selListWidg.count())]
        return [i.text() for i in dateItems if i.checkState()]

    def setEnabled(self, en: bool):
        [i.setEnabled(en) for i in [self.binningCombo, self.saveButton, self.compareDatesButton, self.plotButton]]

    def viewFiles(self):
        self.explorerWindow.refresh()
        self.explorerWindow.show()



# Copyright 2018-2020 Nick Anthony, Backman Biophotonics Lab, Northwestern University
#
# This file is part of PWSpy.
#
# PWSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PWSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PWSpy.  If not, see <https://www.gnu.org/licenses/>.

from __future__ import annotations
import typing
from PyQt5.QtWidgets import QDockWidget, QWidget, \
    QVBoxLayout, QPushButton, QMessageBox, QTabWidget

from .runtimeSettings import AbstractRuntimeAnalysisSettings
from .widgets.QueueAnalysesFrame import AnalysisListItem, QueuedAnalysesFrame
from pwspy_gui.PWSAnalysisApp._dockWidgets.AnalysisSettingsDock.widgets.SettingsFrames import PWSSettingsFrame
from .widgets.SettingsFrames import DynamicsSettingsFrame
from .widgets.SettingsFrames import AbstractSettingsFrame
from pwspy_gui.PWSAnalysisApp.componentInterfaces import CellSelector, AnalysisSettingsCreator


class AnalysisSettingsDock(AnalysisSettingsCreator, QDockWidget):
    def __init__(self, parent: QWidget, cellSelector: CellSelector, erManager):
        super().__init__("Settings", parent=parent)
        self._erManager = erManager
        self.setStyleSheet("QDockWidget > QWidget { border: 1px solid lightgray; }")
        self._selector = cellSelector
        self.setObjectName('AnalysisSettingsDock')  # needed for restore state to work
        widg = QWidget()
        widg.setLayout(QVBoxLayout())
        self._settingsTabWidget = QTabWidget(self)
        self._PWSSettingsFrame = PWSSettingsFrame(self._erManager, cellSelector)
        self._settingsTabWidget.addTab(self._PWSSettingsFrame, "PWS")
        self._DynSettingsFrame = DynamicsSettingsFrame(self._erManager, cellSelector)
        self._settingsTabWidget.addTab(self._DynSettingsFrame, "Dynamics")
        widg.layout().addWidget(self._settingsTabWidget)
        self.addAnalysisButton = QPushButton("Add Analysis")
        widg.layout().addWidget(self.addAnalysisButton)

        self._analysesQueue = QueuedAnalysesFrame(self)

        widg.layout().addWidget(self._analysesQueue)
        self._analysesQueue.setFixedHeight(50)
        widg.setMinimumHeight(200)
        widg.setMinimumWidth(self._PWSSettingsFrame.minimumWidth() + 10)

        self.addAnalysisButton.released.connect(self._addAnalysis)
        self.setWidget(widg)

    def _addAnalysis(self):
        settingsWidget: AbstractSettingsFrame = self._settingsTabWidget.currentWidget()
        try:
            settings = settingsWidget.getSettings()
        except Exception as e:
            QMessageBox.information(self, 'Hold on', str(e))
            return
        self._analysesQueue.addAnalysis(settings)

    def getListedAnalyses(self) -> typing.List[AbstractRuntimeAnalysisSettings]:
        return self._analysesQueue.analyses




import pwspy.dataTypes as pwsdt
import typing as t_
import matplotlib.pyplot as plt
from matplotlib import widgets
import numpy as np

# TODO replace `matplotlib.widgets` with the improved mpl_qt_viz stuff

def selectPointROI(cube: pwsdt.ICBase, side: int = 3, displayIndex: t_.Optional[int] = None):
    """
    Allow the user to select a single point on an image of the acquisition.

    Args:
        side (int): The length (in pixels) of the sides of the square that is used for selection.
        displayIndex (Optional[int]): The z-slice of the 3d array which should be displayed

    Returns:
        np.ndarray: An array of the 4 XY vertices of the square.
    """
    from mpl_qt_viz.roiSelection import PointCreator
    verts = [None]
    if displayIndex is None:
        displayIndex = cube.data.shape[2] // 2
    fig, ax = plt.subplots()
    ax.imshow(cube.data[:, :, displayIndex])
    fig.suptitle("Close to accept ROI")

    def select(Verts, handles):
        verts[0] = Verts

    sel = PointCreator(ax, onselect=select, sideLength=side)
    sel.set_active(True)
    fig.show()
    while plt.fignum_exists(fig.number):
        fig.canvas.flush_events()
    return np.array(verts[0])


def selectLassoROI(cube: pwsdt.ICBase, displayIndex: t_.Optional[int] = None, clim: t_.Sequence = None) -> np.ndarray:
    """
    Allow the user to draw a `freehand` ROI on an image of the acquisition.

    Args:
        displayIndex: Display a particular z-slice of the array for mask drawing. If `None` then the mean along Z is displayed.

    Returns:
        An array of vertices of the polygon drawn.
    """
    Verts = [None]
    if displayIndex is None:
        displayIndex = cube.data.shape[2]//2
    fig, ax = plt.subplots()
    data = cube.data[:, :, displayIndex]
    ax.imshow(data, clim=[np.percentile(data, 1), np.percentile(data, 99)])
    fig.suptitle("Close to accept ROI")

    def onSelect(verts):
        Verts[0] = verts

    l = widgets.LassoSelector(ax, onSelect, lineprops={'color': 'r'})
    fig.show()
    while plt.fignum_exists(fig.number):
        fig.canvas.flush_events()
    return np.array(Verts[0])

def selectRectangleROI(cube: pwsdt.ICBase, displayIndex: t_.Optional[int] = None) -> np.ndarray:
    """
    Allow the user to draw a rectangular ROI on an image of the acquisition.

    Args:
        displayIndex (int): is used to display a particular z-slice for mask drawing. If None then the mean along Z is displayed. Returns an array of vertices of the rectangle.

    Returns:
        np.ndarray: An array of the 4 XY vertices of the rectangle.
    """
    import warnings
    warnings.warn("This method has been moved to the `pwspy_gui.utility` module and will be removed in the future.",
                  category=DeprecationWarning)

    verts = [None]

    if displayIndex is None:
       displayIndex = cube.data.shape[2]//2
    fig, ax = plt.subplots()
    ax.imshow(cube.data[:, :, displayIndex])
    fig.suptitle("Close to accept ROI")

    def rectSelect(mins, maxes):
        verts[0] = ((mins.ydata, mins.xdata), (maxes.ydata, mins.xdata), (maxes.ydata, maxes.xdata), (mins.ydata, maxes.xdata))

    r = widgets.RectangleSelector(ax, rectSelect)
    fig.show()
    while plt.fignum_exists(fig.number):
        fig.canvas.flush_events()
    return np.array(verts[0])



from PyQt5 import QtCore
from PyQt5.QtCore import pyqtSignal
from PyQt5.QtWidgets import QWidget, QGridLayout, QPushButton

from pwspy.utility.acquisition.steps import SequencerStep
from pwspy_gui.PWSAnalysisApp.plugins.acquisitionSequencer._ui.TreeView import MyTreeView, DictTreeView
from pwspy.utility.acquisition.sequencerCoordinate import SequencerCoordinateRange


class SequenceViewer(QWidget):
    newCoordSelected = pyqtSignal(SequencerCoordinateRange)

    def __init__(self, parent: QWidget = None):
        super().__init__(parent, QtCore.Qt.Window)
        self.setWindowTitle("Acquisition Sequence Viewer")

        l = QGridLayout()
        self.setLayout(l)

        self._sequenceTree = MyTreeView(self)

        self._showSettingsButton = QPushButton("Show Settings")
        self._showSettingsButton.released.connect(self._showHideSettings)

        self._selectButton = QPushButton("Update Selection")
        def func():
            self._sequenceTree.commitAndClose() 
            self.newCoordSelected.emit(self._sequenceTree.getCurrentSelectedCoordinateRange())
        self._selectButton.released.connect(func)

        self._settingsTree = DictTreeView()
        self._settingsTree.setColumnCount(2)
        self._settingsTree.setIndentation(10)
        self._sequenceTree.currentItemChanged.connect(lambda item: self._settingsTree.setDict(item.settings))

        l.addWidget(self._sequenceTree, 0, 0)
        l.addWidget(self._selectButton, 1, 0)
        l.addWidget(self._showSettingsButton, 2, 0)
        l.addWidget(self._settingsTree, 0, 1, 1, 3)
        self._settingsTree.hide()

    def setSequenceStepRoot(self, root: SequencerStep):
        self._sequenceTree.setRoot(root)
        self._sequenceTree.expandAll()

    def _showHideSettings(self):
        w = self.width()
        if self._showSettingsButton.text() == "Show Settings":
            self._showSettingsButton.setText("Hide Settings")
            self.setFixedWidth(w*2)
            self._settingsTree.show()
        else:
            self._showSettingsButton.setText("Show Settings")
            self._settingsTree.hide()
            self.setFixedWidth(int(w / 2))